from abc import abstractmethod

class activity_model(object):
	""" The activity model object is for implementing activity models
	for volatile species in melts. It contains all the methods required to
	evaluate the activity.
	"""
	def __init__(self):
		self.set_calibration_ranges([])

	def set_calibration_ranges(self,calibration_ranges):
		self.calibration_ranges = calibration_ranges

	@abstractmethod
	def activity(self,X,**kwargs):
		"""
		"""

	# @abstractmethod
	def check_calibration_range(self,parameters,report_nonexistance=True):
		s = ''
		for cr in self.calibration_ranges:
			if cr.check(parameters) == False:
				s += cr.string(parameters,report_nonexistance)
		return s


# --------------- ACTVITY MODELS ------------------------------- #

class activity_idealsolution(activity_model):
	""" Implements an ideal solution activity model, i.e. it
	will always return the mole fraction.
	"""

	def activity(self,X):
		""" The activity of the component in an ideal solution, i.e., it
		will return the mole fraction.

		Parameters
		----------
		X   float
			The mole fraction of the species in the solution.

		Returns
		-------
		float
			The activity of the species in the solution, i.e., the mole fraction.
		"""
		return X



from VESIcal import core
from VESIcal import calibrations
from VESIcal.tasplot import *

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.pyplot import *
import pandas as pd
import numpy as np
import warnings as w

# ---------- DEFINE CUSTOM PLOTTING FORMATTING ------------ #
style = "seaborn-colorblind"
plt.style.use(style)
plt.rcParams["mathtext.default"] = "regular"
plt.rcParams["mathtext.fontset"] = "dejavusans"
mpl.rcParams['patch.linewidth'] = 1
mpl.rcParams['axes.linewidth'] = 1
plt.rcParams['axes.titlesize'] = 20
plt.rcParams['axes.labelsize'] = 18
plt.rcParams['xtick.labelsize'] = 14
plt.rcParams['ytick.labelsize'] = 14
plt.rcParams['legend.fontsize'] = 14
mpl.rcParams['lines.markersize'] = 10

# Define color cycler based on plot style set here
the_rc = plt.style.library[style] #get style formatting set by plt.style.use()
color_list = the_rc['axes.prop_cycle'].by_key()['color'] * 10 #list of colors by hex code
color_cyler = the_rc['axes.prop_cycle'] #get the cycler

# ----------- MAGMASAT PLOTTING FUNCTIONS ----------- #
def smooth_isobars_and_isopleths(isobars=None, isopleths=None):
	"""
	Takes in a dataframe with calculated isobar and isopleth information (e.g., output from calculate_isobars_and_isopleths)
	and smooths the data for plotting.

	Parameters
	----------
	isobars: pandas DataFrame
		OPTIONAL. DataFrame object containing isobar information as calculated by calculate_isobars_and_isopleths.

	isopleths: pandas DataFrame
		OPTIONAL. DataFrame object containing isopleth information as calculated by calculate_isobars_and_isopleths.

	Returns
	-------
	pandas DataFrame
		DataFrame with x and y values for all isobars and all isopleths. Useful if a user wishes to do custom plotting
		with isobar and isopleth data rather than using the built-in `plot_isobars_and_isopleths()` function.
	"""
	np.seterr(divide='ignore', invalid='ignore') #turn off numpy warning
	w.filterwarnings("ignore", message="Polyfit may be poorly conditioned")

	if isobars is not None:
		P_vals = isobars.Pressure.unique()
		isobars_lists = isobars.values.tolist()
		# add zero values to volatiles list
		isobars_lists.append([0.0, 0.0, 0.0, 0.0])

		isobars_pressure = []
		isobars_H2O_liq = []
		isobars_CO2_liq = []
		# do some data smoothing
		for pressure in P_vals:
			Pxs = [item[1] for item in isobars_lists if item[0] == pressure]
			Pys = [item[2] for item in isobars_lists if item[0] == pressure]

			try:
				## calcualte polynomial
				Pz = np.polyfit(Pxs, Pys, 3)
				Pf = np.poly1d(Pz)

				## calculate new x's and y's
				Px_new = np.linspace(Pxs[0], Pxs[-1], 50)
				Py_new = Pf(Px_new)

				# Save x's and y's
				Px_new_list = list(Px_new)
				isobars_H2O_liq += Px_new_list

				Py_new_list = list(Py_new)
				isobars_CO2_liq += Py_new_list

				pressure_vals_for_list = [pressure]*len(Px_new)
				isobars_pressure += pressure_vals_for_list

			except:
				Px_list = list(Pxs)
				isobars_H2O_liq += Px_list

				Py_list = list(Pys)
				isobars_CO2_liq += Py_list

				pressure_vals_for_list = [pressure]*len(Pxs)
				isobars_pressure += pressure_vals_for_list

		isobar_df = pd.DataFrame({"Pressure": isobars_pressure,
							  "H2O_liq": isobars_H2O_liq,
							  "CO2_liq": isobars_CO2_liq})

	if isopleths is not None:
		XH2O_vals = isopleths.XH2O_fl.unique()
		isopleths_lists = isopleths.values.tolist()

		isopleths_XH2O_fl = []
		isopleths_H2O_liq = []
		isopleths_CO2_liq = []
		for Xfl in XH2O_vals:
			Xxs = [item[1] for item in isopleths_lists if item[0] == Xfl]
			Xys = [item[2] for item in isopleths_lists if item[0] == Xfl]

			try:
				## calcualte polynomial
				Xz = np.polyfit(Xxs, Xys, 2)
				Xf = np.poly1d(Xz)

				## calculate new x's and y's
				Xx_new = np.linspace(Xxs[0], Xxs[-1], 50)
				Xy_new = Xf(Xx_new)

				# Save x's and y's
				Xx_new_list = list(Xx_new)
				isopleths_H2O_liq += Xx_new_list

				Xy_new_list = list(Xy_new)
				isopleths_CO2_liq += Xy_new_list

				XH2Ofl_vals_for_list = [Xfl]*len(Xx_new)
				isopleths_XH2O_fl += XH2Ofl_vals_for_list

			except:
				Xx_list = list(Xxs)
				isopleths_H2O_liq += Xx_list

				Xy_list = list(Xys)
				isopleths_CO2_liq += Xy_list

				XH2Ofl_vals_for_list = [Xfl]*len(Xxs)
				isopleths_XH2O_fl += XH2Ofl_vals_for_list

		isopleth_df = pd.DataFrame({"XH2O_fl": isopleths_XH2O_fl,
							  "H2O_liq": isopleths_H2O_liq,
							  "CO2_liq": isopleths_CO2_liq})

	np.seterr(divide='warn', invalid='warn') #turn numpy warning back on
	w.filterwarnings("always", message="Polyfit may be poorly conditioned")

	if isobars is not None:
		if isopleths is not None:
			return isobar_df, isopleth_df
		else:
			return isobar_df
	else:
		if isopleths is not None:
			return isopleth_df


def plot(isobars=None, isopleths=None, degassing_paths=None, custom_H2O=None, custom_CO2=None,
		 isobar_labels=None, isopleth_labels=None, degassing_path_labels=None, custom_labels=None,
		 custom_colors="VESIcal", custom_symbols=None, markersize=10, figsize=(12,8), save_fig=False,
		 extend_isobars_to_zero=True, smooth_isobars=False, smooth_isopleths=False, **kwargs):
	"""
	Custom automatic plotting of model calculations in VESIcal.
	Isobars, isopleths, and degassing paths can be plotted. Labels can be specified for each.
	Any combination of isobars, isopleths, and degassing paths can be plotted.

	Parameters
	----------
	isobars: pandas DataFrame or list
		OPTIONAL. DataFrame object containing isobar information as calculated by calculate_isobars_and_isopleths. Or a list
		of DataFrame objects.

	isopleths: pandas DataFrame or list
		OPTIONAL. DataFrame object containing isopleth information as calculated by calculate_isobars_and_isopleths. Or a list
		of DataFrame objects.

	degassing_paths: list
		OPTIONAL. List of DataFrames with degassing information as generated by calculate_degassing_path().

	custom_H2O: list
		OPTIONAL. List of groups of H2O values to plot as points. For example myfile.data['H2O'] is one group of H2O values.
		Must be passed with custom_CO2 and must be same length as custom_CO2.

	custom_CO2: list
		OPTIONAL. List of groups of CO2 values to plot as points.For example myfile.data['CO2'] is one group of CO2 values.
		Must be passed with custom_H2O and must be same length as custom_H2O.

	isobar_labels: list
		OPTIONAL. Labels for the plot legend. Default is None, in which case each plotted line will be given the generic
		legend name of "Isobars n", with n referring to the nth isobars passed. Isobar pressure is given in parentheses.
		The user can pass their own labels as a list of strings. If more than one set of isobars is passed, the labels should
		refer to each set of isobars, not each pressure.

	isopleth_labels: list
		OPTIONAL. Labels for the plot legend. Default is None, in which case each plotted isopleth will be given the generic
		legend name of "Isopleth n", with n referring to the nth isopleths passed. Isopleth XH2O values are given in
		parentheses. The user can pass their own labels as a list of strings. If more than one set of isopleths is passed,
		the labels should refer to each set of isopleths, not each XH2O value.

	degassing_path_labels: list
		OPTIONAL. Labels for the plot legend. Default is None, in which case each plotted line will be given the generic
		legend name of "Pathn", with n referring to the nth degassing path passed. The user can pass their own labels
		as a list of strings.

	custom_labels: list
		OPTIONAL. Labels for the plot legend. Default is None, in which case each group of custom points will be given the
		generic legend name of "Customn", with n referring to the nth degassing path passed. The user can pass their own labels
		as a list of strings.

	custom_colors: list
		OPTIONAL. Default value is "VESIcal", which uses VESIcal's color ramp. A list of color values readable by matplotlib
		can be passed here if custom symbol colors are desired. The length of this list must match that of custom_H2O and
		custom_CO2.

	custom_symbols: list
		OPTIONAL. Default value is None, in which case data are plotted as filled circles.. A list of symbol tyles readable
		by matplotlib can be passed here if custom symbol types are desired. The length of this list must match that of
		custom_H2O and custom_CO2.

	markersize: int
		OPTIONAL. Default value is 10. Same as markersize kwarg in matplotlib. Any numeric value passed here will set the
		marker size for (custom_H2O, custom_CO2) points.

	figsize: tuple
		OPTIONAL. Default value is (12,8). Sets the matplotlib.pyplot figsize value as (x_dimension, y_dimension)

	save_fig: False or str
		OPTIONAL. Default value is False, in which case the figure will not be saved. If a string is passed,
		the figure will be saved with the string as the filename. The string must include the file extension.

	extend_isobars_to_zero: bool
		OPTIONAL. If True (default), isobars will be extended to zero, even if there is a finite solubility at zero partial pressure.

	smooth_isobars: bool
		OPTIONAL. Default is False. If set to True, isobar data will be fit to a polynomial and plotted. If False, the raw input data
		will be plotted.

	smooth_isopleths: bool
		OPTIONAL. Default is False. If set to True, isopleth data will be fit to a polynomial and plotted. If False, the raw input data
		will be plotted.

	Returns
	-------
	fig, axes Matplotlib objects
		fig and axes matploblib objects defining a plot with x-axis as H2O wt% in the melt and y-axis as CO2 wt%
		in the melt. Isobars, or lines of constant pressure at which the sample magma composition is saturated,
		and isopleths, or lines of constant fluid composition at which the sample magma composition is saturated,
		are plotted if passed. Degassing paths, or the concentration of dissolved H2O and CO2 in a melt
		equilibrated along a path of decreasing pressure, is plotted if passed.
	"""
	## ------ TURN OFF WARNINGS ------ ##
	np.seterr(divide='ignore', invalid='ignore') #turn off numpy warning
	w.filterwarnings("ignore", message="Polyfit may be poorly conditioned")

	def check_inputs(custom_H2O, custom_CO2):
		if custom_H2O is not None:
			if custom_CO2 is None:
				raise core.InputError("If x data is passed, y data must also be passed.")
			else:
				if len(custom_H2O) == len(custom_CO2):
					pass
				else:
					raise core.InputError("x and y data must be same length")
		if custom_CO2 is not None:
			if custom_H2O is None:
				raise core.InputError("If y data is passed, x data must also be passed.")

	def check_colors(custom_colors):
		if custom_colors == "VESIcal":
			use_colors = color_list
		elif isinstance(custom_colors, list):
			use_colors = custom_colors
		else:
			raise core.InputError("Argument custom_colors must be type list. Just passing one item? Try putting square brackets, [], around it.")
		return use_colors

	def calc_extend_isobars_to_zero(Pxs, Pys):
		"""
		Calculates new end-points for plotting isobars when extend_isobars_to_zero option is set to True.

		Parameters
		----------
		Pxs, Pys: list
			List of x and y values corresponding to isobars.
		"""
		if Pxs[0]*Pys[0] != 0.0:
			if Pxs[0] > Pys[0]:
				Px_new = np.zeros(np.shape(Pxs)[0]+1) #create new array of length n+1 if n is the length of the original array
				Px_new[0] = 0 #set the first x value in the new array equal to 0
				Px_new[1:] = Pxs #fill the rest of the new array with the original array values
				Pxs = Px_new #overwrite the original array with the new one

				Py_new = np.zeros(np.shape(Pys)[0]+1)
				Py_new[0] = Pys[0]
				Py_new[1:] = Pys
				Pys = Py_new
			else:
				Px_new = np.zeros(np.shape(Pxs)[0]+1)
				Px_new[0] = Pxs[0]
				Px_new[1:] = Pxs
				Pxs = Px_new

				Py_new = np.zeros(np.shape(Pys)[0]+1)
				Py_new[0] = 0
				Py_new[1:] = Pys
				Pys = Py_new

		if Pxs[-1]*Pys[-1] != 0.0:
			if Pxs[-1] < Pys[-1]:
				Px_new = np.zeros(np.shape(Pxs)[0]+1)
				Px_new[-1] = 0
				Px_new[:-1] = Pxs
				Pxs = Px_new

				Py_new = np.zeros(np.shape(Pys)[0]+1)
				Py_new[-1] = Pys[-1]
				Py_new[:-1] = Pys
				Pys = Py_new
			else:
				Px_new = np.zeros(np.shape(Pxs)[0]+1)
				Px_new[-1] = Pxs[-1]
				Px_new[:-1] = Pxs
				Pxs = Px_new

				Py_new = np.zeros(np.shape(Pys)[0]+1)
				Py_new[-1] = 0
				Py_new[:-1] = Pys
				Pys = Py_new

		return Pxs, Pys

	## -------- HANDLE USER INPUT ERRORS, SET COLORS, SMOOTH ISOBARS/PLETHS IF DESIRED -------- ##
	check_inputs(custom_H2O=custom_H2O, custom_CO2=custom_CO2)
	use_colors = check_colors(custom_colors=custom_colors)

	if smooth_isobars == True:
		isobars = smooth_isobars_and_isopleths(isobars=isobars)
	if smooth_isopleths == True:
		isopleths = smooth_isobars_and_isopleths(isopleths=isopleths)

	## -------- CREATE FIGURE -------- ##
	fig, ax = subplots(figsize=figsize)
	if 'custom_x' in kwargs:
		ax.set(xlabel=kwargs['xlabel'], ylabel=kwargs['ylabel'])
	else:
		ax.set(xlabel='H$_2$O wt%', ylabel='CO$_2$ wt%')

	labels = []

	## -------- PLOT ISOBARS -------- ##
	if isobars is not None:
		if isinstance(isobars, pd.DataFrame):
			isobars = [isobars]

		for i in range(len(isobars)):
			P_vals = isobars[i].Pressure.unique()
			isobars_lists = isobars[i].values.tolist()

			# add zero values to volatiles list
			isobars_lists.append([0.0, 0.0, 0.0, 0.0])

			P_iter = 0
			for pressure in P_vals:
				P_iter += 1
				Pxs = [item[1] for item in isobars_lists if item[0] == pressure]
				Pys = [item[2] for item in isobars_lists if item[0] == pressure]

				if extend_isobars_to_zero == True:
					try:
						Pxs, Pys = calc_extend_isobars_to_zero(Pxs, Pys)
					except:
						pass
				else:
					print(extend_isobars_to_zero)

				if len(isobars) > 1:
					if P_iter == 1:
						P_list = [int(i) for i in P_vals]
						if isinstance(isobar_labels, list):
							labels.append(str(isobar_labels[i]) + ' (' + ', '.join(map(str, P_list)) + " bars)")
						else:
							labels.append('Isobars ' + str(i+1) + ' (' + ', '.join(map(str, P_list)) + " bars)")
					else:
						labels.append('_nolegend_')

				if len(isobars) > 1:
					ax.plot(Pxs, Pys, color=color_list[i])
				else:
					ax.plot(Pxs, Pys)

			if len(isobars) == 1:
				labels += [str(P_val) + " bars" for P_val in P_vals]

	## -------- PLOT ISOPLETHS -------- ##
	if isopleths is not None:
		if isinstance(isopleths, pd.DataFrame):
			isopleths = [isopleths]

		for i in range(len(isopleths)):
			XH2O_vals = isopleths[i].XH2O_fl.unique()
			isopleths_lists = isopleths[i].values.tolist()

			H_iter = 0
			for Xfl in XH2O_vals:
				H_iter += 1
				Xxs = [item[1] for item in isopleths_lists if item[0] == Xfl]
				Xys = [item[2] for item in isopleths_lists if item[0] == Xfl]

				if len(isopleths) > 1:
					if H_iter == 1:
						H_list = [i for i in XH2O_vals]
						if isinstance(isopleth_labels, list):
							labels.append(str(isopleth_labels[i]) + ' (' + ', '.join(map(str, H_list)) + " XH2Ofluid)")
						else:
							labels.append('Isopleths ' + str(i+1) + ' (' + ', '.join(map(str, H_list)) + " XH2Ofluid)")
					else:
						labels.append('_nolegend_')
					ax.plot(Xxs, Xys, ls='dashed', color=color_list[i])

				if len(isopleths) == 1:
					H_list = [i for i in XH2O_vals]
					if H_iter == 1:
						labels.append('Isopleths (' + ', '.join(map(str, H_list)) + " XH2Ofluid)")
					else:
						labels.append('_nolegend_')
					ax.plot(Xxs, Xys, ls='dashed', color='k')

	## -------- PLOT DEGASSING PATHS -------- ##
	if degassing_paths is not None:
		if isinstance(degassing_paths, pd.DataFrame):
			degassing_paths = [degassing_paths]

		degassing_colors = color_list.copy()
		#degassing_colors.reverse()
		iterno = 0
		for i in range(len(degassing_paths)):
			if degassing_path_labels == None:
				iterno += 1
				labels.append('Path%s' %iterno)
				ax.plot(degassing_paths[i]["H2O_liq"], degassing_paths[i]["CO2_liq"], ls='dotted', color=degassing_colors[i])
			else:
				labels.append(degassing_path_labels[iterno])
				ax.plot(degassing_paths[i]["H2O_liq"], degassing_paths[i]["CO2_liq"], ls='dotted', color=degassing_colors[i])
				iterno += 1

		for i in range(len(degassing_paths)):
			ax.plot(degassing_paths[i]["H2O_liq"].max(), degassing_paths[i]["CO2_liq"].max(), 'o', color=degassing_colors[i])
			labels.append('_nolegend_')

	## -------- PLOT CUSTOM H2O-CO2 -------- ##
	if custom_H2O is not None and custom_CO2 is not None:
		if isinstance(custom_H2O, pd.DataFrame):
			custom_H2O = [custom_H2O]
		if isinstance(custom_CO2, pd.DataFrame):
			custom_CO2 = [custom_CO2]

		if custom_symbols == None:
			use_marker = ['o'] * len(custom_H2O)
		else:
			use_marker = custom_symbols

		iterno = 0
		for i in range(len(custom_H2O)):
			if custom_labels == None:
				iterno +=1
				labels.append('Custom%s' %iterno)
				ax.plot(custom_H2O[i], custom_CO2[i], use_marker[i], color=use_colors[i], markersize=markersize)
			else:
				labels.append(custom_labels[iterno])
				ax.plot(custom_H2O[i], custom_CO2[i], use_marker[i], color=use_colors[i], markersize=markersize)
				iterno += 1

	## -------- PLOT CUSTOM X-Y -------- ##
	if 'custom_x' in kwargs:
			custom_x = kwargs['custom_x']
			custom_y = kwargs['custom_y']

			if isinstance(custom_x, pd.core.series.Series):
				custom_x = [list(custom_x.values)]
			if isinstance(custom_y, pd.core.series.Series):
				custom_y = [list(custom_y.values)]

			if custom_symbols == None:
				use_marker = ['o'] * len(custom_x)
			else:
				use_marker = custom_symbols

			iterno = 0
			for i in range(len(custom_x)):
				if custom_labels == None:
					iterno +=1
					labels.append('Custom%s' %iterno)
					ax.plot(custom_x[i], custom_y[i], use_marker[i], color=use_colors[i], markersize=markersize)
				else:
					labels.append(custom_labels[iterno])
					ax.plot(custom_x[i], custom_y[i], use_marker[i], color=use_colors[i], markersize=markersize)
					iterno += 1

	## -------- PLOT LEGEND -------- ##
	ax.legend(labels, bbox_to_anchor=(1.01,1), loc='upper left')

	if 'custom_x' not in kwargs:
		ax.set_xlim(left=0)
		ax.set_ylim(bottom=0)

	np.seterr(divide='warn', invalid='warn') #turn numpy warning back on
	w.filterwarnings("always", message="Polyfit may be poorly conditioned")

	## -------- SAVE FIGURE IF DESIRED -------- ##
	if save_fig != False:
		fig.savefig(save_fig)

	return fig, ax

def scatterplot(custom_x, custom_y, xlabel=None, ylabel=None, **kwargs):
	"""
	Custom x-y plotting using VESIcal's built-in plot() function, built on Matplotlib's plot and scatter functions.

	Parameters
	----------
	custom_x: list
		List of groups of x-values to plot as points or lines

	custom_y: list
		List of groups of y-values to plot as points or lines

	xlabel: str
		OPTIONAL. What to display along the x-axis.

	ylabel: str
		OPTIONAL. What to display along the y-axis.

	kwargs:
		Can take in any key word agruments that can be passed to `plot()`.

	Returns
	-------
	fig, ax matplotlib objects
		X-y plot with custom x and y axis values and labels.
	"""

	if isinstance(custom_x, list) and isinstance(custom_y, list):
		if len(custom_x) != len(custom_y):
			raise core.InputError("X and y lists must be same length")

	if xlabel is not None:
		if isinstance(xlabel, str):
			pass
		else:
			raise core.InputError("xlabel must be string")

	if ylabel is not None:
		if isinstance(ylabel, str):
			pass
		else:
			raise core.InputError("ylabel must be string")

	return plot(custom_x=custom_x, custom_y=custom_y, xlabel=xlabel, ylabel=ylabel, **kwargs)

# ------- Define custom plotting tools for checking calibrations ------- #

def calib_plot(user_data=None, model='all', plot_type='TAS', zoom=None, figsize=(17,8), legend=True, save_fig=False, **kwargs):
	"""
	Plots user data and calibration set of any or all models on any x-y plot or a total alkalis vs silica (TAS) diagram.
	TAS diagram boundaries provided by tasplot python module, copyright John A Stevenson.

	Parameters
	----------
	user_data: BatchFile object, pandas DataFrame, pandas Series, or dict
		OPTIONAL. Default value is None, in which case only the model calibration set is plotted.
		User provided sample data describing the oxide composition of one or more samples. Multiple samples
		can be passed as an BatchFile object or pandas DataFrame. A single sample can be passed as a pandas
		Series.

	model: str or list
		OPTIONAL. Default value is 'all', in which case all model calibration datasets will be plotted.
		'Mixed' can be used to plot all mixed fluid models.
		String of the name of the model calibration dataset to plot (e.g., 'Shishkina'). Multiple models
		can be plotted by passing them as strings within a list (e.g., ['Shishkina', 'Dixon']).

	plot_type: str
		OPTIONAL. Default value is 'TAS', which returns a total alkali vs silica (TAS) diagram. Any two oxides can
		be plotted as an x-y plot by setting plot_type='xy' and specifying x- and y-axis oxides, e.g., x='SiO2', y='Al2O3'

	zoom: str or list
		OPTIONAL. Default value is None in which case axes will be set to the default of 35<x<100 wt% and 0<y<25 wt% for
		TAS type plots and the best values to show the data for xy type plots. Can pass "user_data" to plot the figure
		where the x and y axes are scaled down to zoom in and only show the region surrounding the user_data. A list of
		tuples may be passed to manually specify x and y limits. Pass in data as  [(x_min, x_max), (y_min, y_max)].
		For example, the default limits here would be passed in as [(35,100), (0,25)].

	figsize: tuple
		OPTIONAL. Default value is (17,8). Sets the matplotlib.pyplot figsize value as (x_dimension, y_dimension)

	legend: bool
		OPTIONAL. Default value is True. Can be set to False in which case the legend will not be displayed.

	save_fig: False or str
		OPTIONAL. Default value is False, in which case the figure will not be saved. If a string is passed,
		the figure will be saved with the string as the filename. The string must include the file extension.

	Returns
	-------
	matplotlib object
	"""

	#Get x and y axis limits, if user passed them
	if zoom == None:
		user_xmin = 35
		user_xmax = 100
		user_ymin = 0
		user_ymax = 25
	elif zoom == 'user_data':
		if isinstance(user_data, pd.DataFrame):
			print("'user_data' type zoom for more than one sample is not implemented yet.")
			user_xmin = 35
			user_xmax = 100
			user_ymin = 0
			user_ymax = 25
		elif isinstance(user_data, pd.core.series.Series) or isinstance(user_data, dict):
			user_xmin = user_data['SiO2'] - 5
			user_xmax = user_data['SiO2'] + 5
			user_ymin = user_data['Na2O'] + user_data['K2O'] - 2
			if user_ymin <0:
				user_ymin = 0
			user_ymax = user_data['Na2O'] + user_data['K2O'] + 2
	elif isinstance(zoom, list):
		user_xmin, user_xmax = zoom[0]
		user_ymin, user_ymax = zoom[1]

	#Create the figure
	fig, ax1 = plt.subplots(figsize = figsize)
	font = {'family': 'sans-serif',
				'color':  'black',
				'weight': 'normal',
				'size': 20,
				}

	#TAS figure
	if plot_type == 'TAS':
		ax1.set_xlim([user_xmin, user_xmax]) # adjust x limits here if you want to focus on a specific part of compostional space
		ax1.set_ylim([user_ymin, user_ymax]) # adjust y limits here
		plt.xlabel('SiO$_2$, wt%', fontdict=font, labelpad = 15)
		plt.ylabel('Na$_2$O+K$_2$O, wt%', fontdict=font, labelpad = 15)
		if zoom == None:
			add_LeMaitre_fields(ax1)
	elif plot_type == 'xy':
		if 'x' in kwargs and 'y' in kwargs:
			x = kwargs['x']
			y = kwargs['y']
			if zoom != None:
				ax1.set_xlim([user_xmin, user_xmax])
				ax1.set_ylim([user_ymin, user_ymax])
			plt.xlabel(str(x)+", wt%", fontdict=font, labelpad = 15)
			plt.ylabel(str(y)+", wt%", fontdict=font, labelpad = 15)
		else:
			raise core.InputError("If plot_type is 'xy', then x and y values must be passed as strings. For example, x='SiO2', y='Al2O3'.")

	#Plot Calibration Data
	if model == 'all':
		model = ['MagmaSat',
				'Shishkina',
				'Dixon',
				'IaconoMarziano',
				'Liu',
				#'EguchiCarbon',
				'AllisonCarbon',
				'MooreWater']
	if model == 'mixed':
		model = ['MagmaSat',
				 'Shishkina',
				 'Dixon',
				 'IaconoMarziano',
				 'Liu']

	if isinstance(model, str):
		model = [model]

	if isinstance(model, list):
		for modelname in model:
			model_type = calibrations.return_calibration_type(modelname)
			if model_type['H2O'] == True:
				h2o_legend = True
			if model_type['CO2'] == True or model_type['Mixed'] == True:
				co2_h2oco2_legend = True

		if h2o_legend == True:
			plt.scatter([], [], marker='', label=r"$\bf{Pure \ H_2O:}$")

			for modelname in model:
				calibdata = calibrations.return_calibration(modelname)
				model_type = calibrations.return_calibration_type(modelname)
				if isinstance(calibdata, str):
					w.warn(calibdata)
				else:
					if model_type['H2O'] == True:
						if plot_type == 'TAS':
							try:
								plt.scatter(calibdata['H2O']['SiO2'], calibdata['H2O']['Na2O'] + calibdata['H2O']['K2O'],
											marker='s', facecolors=calibdata['facecolor'], edgecolors='k', label=str(modelname))
							except:
								plt.scatter(calibdata['H2O']['SiO2'], calibdata['H2O']['Na2O+K2O'],
											marker='s', facecolors=calibdata['facecolor'], edgecolors='k', label=str(modelname))
						if plot_type == 'xy':
							try:
								plt.scatter(calibdata['H2O'][x], calibdata['H2O'][y],
											marker='s', facecolors=calibdata['facecolor'], edgecolors='k', label=str(modelname))
							except:
								w.warn("The requested oxides were not found in the calibration dataset for " + str(modelname) + ".")

			if co2_h2oco2_legend == True:
				plt.scatter([], [], marker='', label=r"${\ }$")

		if co2_h2oco2_legend == True:
			plt.scatter([], [], marker='', label=r"$\bf{\ CO_2 \ and \ H_2O\!-\!CO_2:}$")

		for modelname in model:
			calibdata = calibrations.return_calibration(modelname)
			model_type = calibrations.return_calibration_type(modelname)
			if isinstance(calibdata, str):
				w.warn(calibdata)
			else:
				if model_type['CO2'] == True and model_type['Mixed'] == True:
					frames = [calibdata['CO2'], calibdata['Mixed']]
					co2_and_mixed = pd.concat(frames)
					if plot_type == 'TAS':
						try:
							plt.scatter(co2_and_mixed['SiO2'], co2_and_mixed['Na2O'] + co2_and_mixed['K2O'],
										marker='d', facecolors=calibdata['facecolor'], edgecolors='k', label=str(modelname))
						except:
							plt.scatter(co2_and_mixed['SiO2'], co2_and_mixed['Na2O+K2O'],
									marker='d', facecolors=calibdata['facecolor'], edgecolors='k', label=str(modelname))
					if plot_type == 'xy':
						try:
							plt.scatter(co2_and_mixed[x], co2_and_mixed[y],
									marker='d', facecolors=calibdata['facecolor'], edgecolors='k', label=str(modelname))
						except:
							w.warn("The requested oxides were not found in the calibration dataset for " + str(modelname) + ".")
				elif model_type['CO2'] == True or model_type['Mixed'] == True:
					if model_type['CO2'] == True:
						thistype = 'CO2'
					if model_type['Mixed'] == True:
						thistype = 'Mixed'
					if plot_type == 'TAS':
						try:
							plt.scatter(calibdata[thistype]['SiO2'], calibdata[thistype]['Na2O'] + calibdata[thistype]['K2O'],
										marker='d', facecolors=calibdata['facecolor'], edgecolors='k', label=str(modelname))
						except:
							plt.scatter(calibdata[thistype]['SiO2'], calibdata[thistype]['Na2O+K2O'],
									marker='d', facecolors=calibdata['facecolor'], edgecolors='k', label=str(modelname))
					if plot_type == 'xy':
						try:
							plt.scatter(calibdata[thistype][x], calibdata[thistype][y],
									marker='d', facecolors=calibdata['facecolor'], edgecolors='k', label=str(modelname))
						except:
							w.warn("The requested oxides were not found in the calibration dataset for " + str(modelname) + ".")
	else:
		raise core.InputError("model must be of type str or list")

	#Plot user data
	if user_data is None:
		pass
	else:
		if (user_data.__class__.__module__, user_data.__class__.__name__) == ('VESIcal', 'BatchFile'):
			user_data = user_data.data
			#batchfile and VESIcal (__init__) are not imported to avoid circular imports
			#use above notation to interrogate datatype
		if plot_type == 'TAS':
			_sample = user_data.copy()
			try:
				_sample["TotalAlkalis"] = _sample["Na2O"] + _sample["K2O"]
			except:
				core.InputError("Na2O and K2O data must be in user_data")
			plt.scatter(_sample['SiO2'], _sample['TotalAlkalis'],
						s=150, edgecolors='w', facecolors='red', marker='P',
						label = 'User Data')
		if plot_type == 'xy':
			_sample = user_data.copy()
			plt.scatter(_sample[x], _sample[y],
						s=150, edgecolors='w', facecolors='red', marker='P',
						label = 'User Data')

	if legend == True:
		plt.legend(bbox_to_anchor=(1.04,1), loc="upper left")
	fig.tight_layout()
	if isinstance(save_fig, str):
		fig.savefig(save_fig)

	return plt.show()



import pandas as pd
import os
import sys
import warnings as w

from VESIcal import core
from VESIcal import sample_class

def rename_duplicates(df, suffix='-duplicate-'):
    appendents = (suffix + df.groupby(level=0).cumcount().astype(str).replace('0','')).replace(suffix, '')
    return df.set_index(df.index.astype(str) + appendents)

class status_bar(object):
    """Various styles of status bars that display the progress of a calculation
    within a loop
    """
    def __init__():
        pass

    def status_bar(percent, sample_name=None, btext=None, barLen=20):
        """
        Prints an updating status bar to the terminal or jupyter notebook.

        Parameters
        ----------
        percent: float
            Percent value of progress from 0 to 1

        sample_name: string
            Name of the current sample being calculated

        btext: string
            Any extra text to display next to status bar

        barLen: int
            Length of bar to print
        """
        sys.stdout.write("\r")
        sys.stdout.write("[{:<{}}] {:.0f}%".format("=" * int(barLen * percent), barLen, percent * 100))

        sample_string = str(sample_name)
        # Set max number of characters in sample name
        max_name_length = 25
        if len(str(sample_name)) >= max_name_length:
            sample_string = str(sample_name)[0:max_name_length-1] + "..."

        # Write out sample name and trailing spaces to cover contents of previous sample names left over on line
        if sample_name != None:
            sys.stdout.write("  Working on sample " + sample_string + "                            ")
        if btext != None:
            sys.stdout.write(" " + str(btext))
        if percent == 1.0:
            sys.stdout.write("\n")
        sys.stdout.flush()

# ---------- BATCHFILE CLASS --------- #
class BatchFile(object):
    """A batch file with sample names and oxide compositions

    Attributes
    ----------
        filename: str
            Path to the batch file, e.g., "my_file.xlsx". This always needs to be passed, even if the user is passing a pandas DataFrame
            rather than an batch file. If passing a DataFrame, filename should be set to None. File can be excel file (.xlsx) or .csv.

        sheet_name: str
            OPTIONAL. For Excel files. Default value is 0 which gets the first sheet in the batch spreadsheet file. This implements the pandas.
            read_excel() sheet_name parameter. But functionality to read in more than one sheet at a time (e.g., pandas.read_excel(sheet_name=None))
            is not yet imlpemented in VESIcal. From the pandas 1.0.4 documentation:
            
            Available cases:
            - Defaults to 0: 1st sheet as a DataFrame
            - 1: 2nd sheet as a DataFrame
            - "Sheet1": Load sheet with name “Sheet1”

        file_type: str
            OPTIONAL. Default is 'excel', which denotes that passed file has extension .xlsx. Other option is 'csv', which denotes that
            the passed file has extension .csv.

        units: str
            OPTIONAL. Default is 'wtpt_oxides'. String defining whether the oxide composition is given in wt percent
            ("wtpt_oxides", which is the default), mole oxides (mol_oxides) or mole cations (mol_cations).

        default_normalization:     None or str
            The type of normalization to apply to the data by default. One of:
            - None (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%, including volatiles.
            The volatile wt% will remain fixed, whilst the other major element oxides are reduced
            proportionally so that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to 100%, assuming it is
            volatile-free. If H2O or CO2 are passed to the function, their un-normalized values will
            be retained in addition to the normalized non-volatile oxides, summing to >100%.

        default_units     str
            The type of composition to return by default, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations

        label: str
            OPTIONAL. Default is 'Label'. Name of the column within the passed file referring to sample names.

        dataframe: pandas DataFrame
            OPTIONAL. Default is None in which case this argument is ignored. This argument is used when the user wishes to turn
            a pandas DataFrame into an BatchFile object, for example when user data is already in python rather than being imported
            from a file. In this case set `dataframe` equal to the dataframe object being passed in. If using this option, pass
            None to filename.
    """
    def __init__(self, filename, sheet_name=0, file_type='excel', units='wtpt_oxides', label='Label', default_normalization='none', default_units='wtpt_oxides', dataframe=None, **kwargs):
        """Return a BatchFile object whose parameters are defined here."""
        self.units = units
        self.set_default_normalization(default_normalization)
        self.set_default_units(default_units)

        if filename != None:
            file_name, file_extension = os.path.splitext(filename)
            if file_extension == '.xlsx' or file_extension == '.xls':
                file_type = 'excel'
            if file_extension == '.csv':
                file_type = 'csv'

        if isinstance(sheet_name, str) or isinstance(sheet_name, int):
            pass
        else:
            raise core.InputError("If sheet_name is passed, it must be of type str or int. Currently, VESIcal cannot import more than one sheet at a time.")

        if dataframe is not None:
            data = dataframe
            if label is not None:
                data = self.try_set_index(data, label)
        else:
            if file_type == 'excel':
                data = pd.read_excel(filename, sheet_name=sheet_name)
                data = self.try_set_index(data, label)
            elif file_type == 'csv':
                data = pd.read_csv(filename)
                data = self.try_set_index(data, label)
            else:
                raise core.InputError("file_type must be one of \'excel\' or \'csv\'.")

        data = rename_duplicates(data) #handle any duplicated sample names
        data = data.dropna(how='all') #drop any rows that are all NaNs
        data = data.fillna(0) #fill in any missing data with 0's

        if 'model' in kwargs:
            w.warn("You don't need to pass a model here, so it will be ignored. You can specify a model when performing calculations on your dataset \
(e.g., calculate_dissolved_volatiles())",RuntimeWarning,stacklevel=2)

        if 'norm' in kwargs:
            w.warn("We noticed you passed a norm argument here. This does nothing. You can normalize your BatchFile and save it to a new variable name \
after import using normalize(BatchFileObject). See the documentation for more info.",RuntimeWarning,stacklevel=2)

        total_iron_columns = ["FeOt", "FeOT", "FeOtot", "FeOtotal", "FeOstar", "FeO*"]
        for name in total_iron_columns:
            if name in data.columns:
                if 'FeO' in data.columns:
                    for row in data.itertuples():
                        if data.at[row.Index, "FeO"] == 0 and data.at[row.Index, name] > 0:
                            w.warn("Sample " + str(row.Index) + ": " + str(name) + " value of " + str(data.at[row.Index, name]) + " used as FeO. Fe2O3 set to 0.0.",
                                    RuntimeWarning,stacklevel=2)
                            data.at[row.Index, "Fe2O3"] = 0.0
                            data.at[row.Index, "FeO"] = data.at[row.Index, name]
                else:
                    w.warn("Total iron column " + str(name) + " detected. This column will be treated as FeO. If Fe2O3 data are not given, Fe2O3 will be 0.0. \
In future, an option to calcualte FeO/Fe2O3 based on fO2 will be implemented.",RuntimeWarning,stacklevel=2)
                    data['FeO'] = data[name]

        if units == "wtpt_oxides":
            pass
        if units == "mol_oxides":
            data = self._molOxides_to_wtpercentOxides(data)
        if units == "mol_cations":
            data = self._molCations_to_wtpercentOxides(data)

        for oxide in core.oxides:
            if oxide in data.columns:
                pass
            else:
                data[oxide] = 0.0

        self.data = data

    def set_default_normalization(self, default_normalization):
        """ Set the default type of normalization to use with the get_composition() method.

        Parameters
        ----------
        default_normalization:    str
            The type of normalization to apply to the data. One of:
            - 'none' (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%, including volatiles.
            The volatile wt% will remain fixed, whilst the other major element oxides are reduced
            proportionally so that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to 100%, assuming it is
            volatile-free. If H2O or CO2 are passed to the function, their un-normalized values will
            be retained in addition to the normalized non-volatile oxides, summing to >100%.

        """
        if default_normalization in ['none','standard','fixedvolatiles','additionalvolatiles']:
            self.default_normalization = default_normalization
        else:
            raise core.InputError("The normalization method must be one of 'none', 'standard', 'fixedvolatiles',\
             or 'additionalvolatiles'.")

    def set_default_units(self, default_units):
        """ Set the default units of composition to return when using the get_composition() method.

        Parameters
        ----------
        default_units     str
            The type of composition to return, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations
        """
        if default_units in ['wtpt_oxides','mol_oxides','mol_cations']:
            self.default_units = default_units
        else:
            raise core.InputError("The units must be one of 'wtpt_oxides','mol_oxides','mol_cations'.")

    def get_composition(self, species=None, normalization=None, units=None, exclude_volatiles=False, asBatchFile=False):
        """ Returns a pandas DataFrame containing the compositional information for all samples
        in the BatchFile object

        Parameters
        ----------
        species:    NoneType or str
            The name of the oxide or cation to return the concentration of. If NoneType (default) the
            whole composition of each sample will be returned. If an oxide is passed, the value in
            wtpt will be returned unless units is set to 'mol_oxides', even if the default units for the
            sample object are mol_oxides. If an element is passed, the concentration will be returned as
            mol_cations, unless 'mol_singleO' is specified as units, even if the default units for the
            sample object are mol_singleO. Unless normalization is specified in the method call, none
            will be applied.

        normalization:     NoneType or str
            The type of normalization to apply to the data. One of:
            - 'none' (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%, including volatiles.
            The volatile wt% will remain fixed, whilst the other major element oxides are reduced
            proportionally so that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to 100%, assuming it is
            volatile-free. If H2O or CO2 are passed to the function, their un-normalized values will
            be retained in addition to the normalized non-volatile oxides, summing to >100%.

            If NoneType is passed the default normalization option will be used (self.default_normalization).

        units:     NoneType or str
            The units of composition to return, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations
            - mol_singleO
            If NoneType is passed the default units option will be used (self.default_type).

        exclude_volatiles   bool
            If True, volatiles will be excluded from the returned composition, prior to normalization and
            conversion.

        asBatchFile:    bool
            If True, returns a BatchFile object. If False, returns a pandas.DataFrame object.

        Returns
        -------
        pandas.DataFrame or BatchFile object
            All sample information.
        """
        data = self.data.copy()

        # Fetch the default return types if not specified in function call
        if normalization == None and species == None:
            normalization = self.default_normalization
        if units == None and species == None:
            units = self.default_units

        new_compositions = []
        sample_names = []
        for index, row in data.iterrows():
            sample_comp = self.get_sample_composition(index, units=units, asSampleClass=True)
            new_compositions.append(sample_comp.get_composition(species=species, normalization=normalization, units=units, exclude_volatiles=exclude_volatiles))
            sample_names.append(index)
        if isinstance(new_compositions[0], pd.Series):
            return_frame = pd.concat([pd.DataFrame(j) for j in new_compositions], axis=1)
            return_frame = return_frame.transpose()
            return_frame["new_index"] = sample_names
            return_frame = return_frame.set_index("new_index")
            return_frame.index.name = None
        elif isinstance(new_compositions[0], float):
            species_data = {species: new_compositions}
            return_frame = pd.DataFrame(species_data, index=[name for name in sample_names])
        else:
            return_frame = None

        if asBatchFile == False:
            return return_frame
        else:
            return BatchFile(filename=None, dataframe=return_frame, label=None)

    def get_data(self, normalization=None, units=None, asBatchFile=False):
        """
        Returns all data stored in a BatchFile object (both compositional and other data). To return only the
        compositional data, use get_composition().

        Parameters
        ----------
        normalization:     NoneType or str
            The type of normalization to apply to the data. One of:
            - 'none' (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%, including volatiles.
            The volatile wt% will remain fixed, whilst the other major element oxides are reduced
            proportionally so that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to 100%, assuming it is
            volatile-free. If H2O or CO2 are passed to the function, their un-normalized values will
            be retained in addition to the normalized non-volatile oxides, summing to >100%.

            If NoneType is passed the default normalization option will be used (self.default_normalization).

        units:     NoneType or str
            The units of composition to return, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations
            - mol_singleO

            If NoneType is passed the default units option will be used (self.default_type).

        asBatchFile:    bool
            If True, returns a BatchFile object. If False, returns a pandas.DataFrame object.

        Returns
        -------
        pandas.DataFrame or BatchFile object
            All sample information.
        """
        data = self.data.copy()

        # Fetch the default return units if not specified in function call
        if units == None:
            units = self.default_units

        # Grab all compositional data
        compositional_data = self.get_composition(normalization=normalization, units=units)

        # Grab all non-compositional data
        non_compositional_data = data.filter([col for col in data.columns if col not in core.oxides])

        # concatenate both compositional and non-compositional dataframes into one
        return_frame = pd.concat([compositional_data, non_compositional_data], axis=1)

        if asBatchFile == False:
            return return_frame
        else:
            return BatchFile(filename=None, dataframe=return_frame, label=None)

    def get_sample_composition(self, samplename, species=None, normalization=None, units=None, asSampleClass=False):
        """
        Returns oxide composition of a single sample from a user-imported file as a dictionary

        Parameters
        ----------
        samplename: string
            Name of the desired sample

        normalization: NoneType or str
            The type of normalization to apply to the data. One of:
            - 'none' (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%, including volatiles.
            The volatile wt% will remain fixed, whilst the other major element oxides are reduced
            proportionally so that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to 100%, assuming it is
            volatile-free. If H2O or CO2 are passed to the function, their un-normalized values will
            be retained in addition to the normalized non-volatile oxides, summing to >100%.

            If NoneType is passed the default normalization option will be used (self.default_normalization).

        units:     NoneType or str
            The units of composition to return, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations
            - mol_singleO
            If NoneType is passed the default units option will be used (self.default_type).

        asSampleClass:  bool
            If True, the sample composition will be returned as a sample class, with default options. In this case
            any normalization instructions will be ignored.

        Returns
        -------
        dictionary, float, or sample_class.Sample object
            Composition of the sample as oxides
        """
        # Fetch the default return types if not specified in function call
        if normalization == None and species == None:
            normalization = self.default_normalization
        if units == None and species == None:
            units = self.default_units

        # Check that normalization being chosen is one of the possible options
        if normalization in [None, 'none','standard','fixedvolatiles','additionalvolatiles']:
            pass
        else:
            raise core.InputError("The normalization method must be one of 'none', 'standard', 'fixedvolatiles',\
             or 'additionalvolatiles'.")

        data = self.data
        my_sample = pd.DataFrame(data.loc[samplename])
        sample_dict = (my_sample.to_dict()[samplename])
        sample_oxides = {}
        for item, value in sample_dict.items():
            if item in core.oxides:
                sample_oxides.update({item: value})

        _sample = sample_class.Sample(sample_oxides)

        if asSampleClass == True:
            return _sample
        else:
            return_sample = _sample.get_composition(species=species, units=units, normalization=normalization)
            if species == None:
                return dict(return_sample)
            elif isinstance(species, str):
                return return_sample

    def _molOxides_to_wtpercentOxides(self, data):
        for i, row in data.iterrows():
            sample_comp = {}
            for oxide in core.oxides:
                if oxide in data.columns:
                    sample_comp[oxide] = row[oxide]
                else:
                    sample_comp[oxide] = 0.0
            _sample = sample_class.Sample(sample_comp, units='mol_oxides')
            _sample_conv = _sample.get_composition()
            for ox in core.oxides:
                data.loc[i,oxide] = _sample_conv[oxide]
        return data

    def _molCations_to_wtpercentOxides(self, data):
        for i, row in data.iterrows():
            sample_comp = {}
            for cation in core.oxides_to_cations[core.oxides]:
                if cation in data.columns:
                    sample_comp[cation] = row[cation]
                else:
                    sample_comp[cation] = 0.0
            _sample = sample_class.Sample(sample_comp, units='mol_cations')
            _sample_conv = _sample.get_composition()
            for oxide in core.oxides:
                data.loc[i,oxide] = _sample_conv[oxide]
        return data

    def try_set_index(self, dataframe, label):
        """
        Method to handle setting the index column in an BatchFile object. If no column is passed that matches the default index name,
        then this method will attempt to choose the 'best' column that the user might want to serve as an index column.

        Parameters
        ----------
        dataframe: pandas DataFrame

        label: str
            Name of the column within the passed Excel file referring to sample names.
        """
        _dataframe = dataframe.copy()
        try:
            _dataframe = _dataframe.set_index(label)
        except:
            label_found = False
            for col in _dataframe.columns:
                if col in core.oxides:
                    pass
                else:
                    _dataframe = _dataframe.set_index(col)
                    label_found = True
                    w.warn("No Label column given, so column '" + str(col) + "' was chosen for you. To choose your own, set label='<column-name>'.",RuntimeWarning,stacklevel=2)
                    break
            if label_found == False:
                _dataframe.index.name = 'Label'
                w.warn("No Label column given, so one was created for you. To choose your own, set label='<column-name>'.",RuntimeWarning,stacklevel=2)

        return _dataframe

    def preprocess_sample(self,sample):
        """
        Adds 0.0 values to any oxide data not passed.

        Parameters
        ----------
        sample: pandas DataFrame
            self.data composition of samples in wt% oxides

        Returns
        -------
        pandas DataFrame
        """
        for oxide in core.oxides:
            if oxide in self.data.columns:
                pass
            else:
                self.data[oxide] = 0.0

        return sample

    def save_excel(self, filename, calculations, sheet_names=None):
        """
        Saves data calculated by the user in batch processing mode (using the BatchFile class methods) to an organized
        Excel file, with the original user data plus any calculated data.

        Parameters
        ----------
        filename: string
            Name of the file. Extension (.xlsx) should be passed along with the name itself, all in quotes (e.g., 'myfile.xlsx').

        calculations: pandas DataFrame or list of pandas DataFrames
            A single DataFrame or list of DataFrames (e.g., calculated outputs from any of the core BatchFile functions:
            calculate_dissolved_volatiles, calculate_equilibrium_fluid_comp, and calculate_saturation_pressure). If None,
            only the original user data will be saved.

        sheet_names: None, string, or list
            OPTIONAL. Default value is None. Allows user to set the name of the sheet or sheets written to the Excel file.

        Returns
        -------
            Creates and saves an Excel file with data from each calculation saved to its own sheet.
        """
        if isinstance(calculations, list):
            if isinstance(sheet_names, list) or sheet_names is None:
                pass
            else:
                raise core.InputError("If calculations is passed as list, sheet_names must also be list of same length")
        elif calculations == None:
            pass
        else:
            calculations = [calculations]

        with pd.ExcelWriter(filename) as writer:
            self.data.to_excel(writer, 'Original_User_Data')
            if isinstance(calculations, list):
                if sheet_names is None:
                    for n, df in enumerate(calculations):
                        df.to_excel(writer, 'Calc%s' % n)
                elif isinstance(sheet_names, list):
                    pass
                else:
                    sheet_names = [sheet_names]
                if isinstance(sheet_names, list):
                    if len(sheet_names) == len(calculations):
                        pass
                    else:
                        raise core.InputError("calculations and sheet_names must have the same length")

                    for i in range(len(calculations)):
                        if isinstance(sheet_names[i], str):
                            calculations[i].to_excel(writer, sheet_names[i])
                        else:
                            raise core.InputError("if sheet_names is passed, it must be list of strings")
            elif calculations == None:
                pass
        return print("Saved " + str(filename))

    def save_csv(self, filenames, calculations, **kwargs):
        """
        Saves data calculated by the user in batch processing mode to a comma-separated values (csv) file. Mirros the pandas.to_csv()
        method. Any argument that can be passed to pandas.csv() can be passed here. One csv file will be saved for each calculation
        passed.

        Parameters
        ----------
        filenames: string or list of strings
            Name of the file. Extension (.csv) should be passed along with the name itself, all in quotes (e.g., 'myfile.csv'). The number
            of calculations passed must match the number of filenames passed. If passing more than one, should be passed as a list.

        calculations: pandas DataFrame or list of pandas DataFrames
            A single variable or list of variables containing calculated outputs from any of the core BatchFile functions:
            calculate_dissolved_volatiles, calculate_equilibrium_fluid_comp, and calculate_saturation_pressure.

        Returns
        -------
            Creates and saves a CSV file or files with data from each calculation saved to its own file.
        """
        if type(filenames) != list:
            filenames = [filenames]
        if type(calculations) != list:
            calculations = [calculations]
        if len(filenames) != len(calculations):
            raise core.InputError("calculations and filenames must have the same length")

        for i in range(len(filenames)):
            calculations[i].to_csv(filenames[i], **kwargs)
            print ("Saved " + str(filenames[i]))

def from_DataFrame(dataframe, units='wtpt_oxides', label='Label'):
    """
    Transforms any pandas DataFrame object into a VESIcal BatchFile object.

    Parameters
    ----------
    dataframe: pd.DataFrame object
        DataFrame object containing samples and oxide compositions.

    units: str
        OPTIONAL. Default is 'wtpt_oxides'. String defining whether the oxide composition is given in wt percent
        ("wtpt_oxides", which is the default), mole fraction oxides ("mol_oxides"), or mole fraction cations ("mol_cations").

    label: str
        OPTIONAL. Default is 'Label'. Name of the column within the passed file referring to sample names. This
        column will be set as the index column.

    Returns
    -------
    VESIcal.BatchFile object
    """
    return BatchFile(filename=None, dataframe=dataframe, units=units, label=label)



from abc import abstractmethod
import warnings as w
import numpy as np

from VESIcal import core
from VESIcal import models
from VESIcal.models import magmasat

from copy import deepcopy

class Calculate(object):
    """ The Calculate object is a template for implementing user-friendly methods for
    running calculations using the volatile solubility models. All Calculate methods
    have a common workflow- sample is read in, preprocessed, the calculation is performed,
    the calibration range is checked, and the results stored.
    """
    def __init__(self,sample,model='MagmaSat',silence_warnings=False,**kwargs):
        """
        Initializes the calculation.

        Parameters
        ----------
        sample:    Sample class
            The rock composition as a Sample object.
        model:     string or Model class
            Which model to use for the calculation. If passed a string, it will look up the name in the
            default_models dictionary. Default is MagmaSat.
        silence_warnings:     bool
            Silence warnings about calibration ranges. Default is False.
        preprocess_sample:     bool
            Before running the calculation, run the sample through the preprocessing routine. As of Feb 2021
            this functionality should be redundant.
        """
        self.model_name = model
        if model == 'MagmaSat':
            self.model = magmasat.MagmaSat()
        elif type(model) == str:
            if model in models.default_models.keys():
                self.model = models.default_models[model]
            else:
                raise core.InputError("The model name given is not recognised. Run the method get_model_names() to find allowed names.")
        else:
            self.model = model

        self.sample = sample

        self.result = self.calculate(sample=self.sample,**kwargs)
        self.calib_check = self.check_calibration_range(sample=self.sample,**kwargs)

        if self.calib_check is not None and silence_warnings == False:
            if self.calib_check != '':
                w.warn(self.calib_check,RuntimeWarning)

    @abstractmethod
    def calculate(self):
        """ """

    @abstractmethod
    def check_calibration_range(self):
        """ """

class calculate_dissolved_volatiles(Calculate):
    """ Calculates the dissolved volatile concentration using a chosen model (default is MagmaSat).
    Using this interface will preprocess the sample, run the calculation, and then check
    the calibration ranges. All parameters required by the chosen model must be passed.

    Parameters
    ----------
    sample:    Sample class
        The rock composition as a Sample object.
    pressure:   float
        Total pressure in bars.
    model:  string or Model object
        Model to be used. If using one of the default models, this can be
        the string corresponding to the model in the default_models dict.
    silence_warnings     bool
        If set to True, no warnings will be raised automatically when calibration checks fail.
    preprocess_sample     bool
        If True (default), the sample will be preprocessed according to the preprocessing operations within
        the models. If you obtain unexpected results, try setting to False.

    Returns
    -------
    Calculate object
        Calculate object, access results by fetching the result property. Dissolved
        volatile concentrations (in wt%), in order (CO2, H2O, if using a mixed fluid
        default model).
    """
    def return_default_units(self, sample, calc_result, **kwargs):
            """ Checkes the default units set for the sample_class.Sample object and returns the
            result of a calculation in those units.

            Parameters
            ----------
            sample: Sample class
                The rock composition as a Sample object.

            calc_result: dict or float
                Result of a calculate_dissolved_volatiles() calculation on a sample.
            """
            default_units = sample.default_units

            # get the composition of
            bulk_comp = deepcopy(sample)

            # check if calculation result is H2O-only, CO2-only, or mixed H2O-CO2
            if isinstance(self.model_name, str):
                if self.model_name in models.get_model_names(model='mixed'):
                    # set dissolved H2O and CO2 values. 
                    # this action assumes they are input as wt% but updates the composition in its default units.
                    bulk_comp.change_composition({'H2O': calc_result['H2O_liq'], 
                                                  'CO2': calc_result['CO2_liq']})
                    return {'H2O_liq': bulk_comp.get_composition(species='H2O', units=default_units),
                            'CO2_liq': bulk_comp.get_composition(species='CO2', units=default_units)}
                elif 'Water' in self.model_name:
                    bulk_comp.change_composition({'H2O': calc_result})
                    return bulk_comp.get_composition(species='H2O', units=default_units)
                elif 'Carbon' in self.model_name:
                    bulk_comp.change_composition({'CO2': calc_result})
                    return bulk_comp.get_composition(species='CO2', units=default_units)
                elif self.model_name == 'MagmaSat':
                    bulk_comp.change_composition({'H2O': calc_result['H2O_liq'], 
                                                  'CO2': calc_result['CO2_liq']})
                    if 'verbose' in kwargs and kwargs['verbose'] == True: # check if verbose method has been chosen
                        return {'H2O_liq': bulk_comp.get_composition(species='H2O', units=default_units),
                                'CO2_liq': bulk_comp.get_composition(species='CO2', units=default_units),
                                'XH2O_fl': calc_result['XH2O_fl'],
                                'XCO2_fl': calc_result['XCO2_fl'],
                                'FluidProportion_wt': calc_result['FluidProportion_wt']}
                    else:
                        return {'H2O_liq': bulk_comp.get_composition(species='H2O', units=default_units),
                                'CO2_liq': bulk_comp.get_composition(species='CO2', units=default_units)}
            else: 
            # if self.model_name is not a string, most likely this is a user-created model class, 
            # and so we cannot interrogate it for model type (H2O, CO2, or mixed)
            # TODO: create model type parameter so that we can ID model type this way instead of via strings in a list!
                return calc_result

    def calculate(self,sample,pressure,**kwargs):
        dissolved = self.model.calculate_dissolved_volatiles(pressure=pressure,sample=sample,returndict=True,**kwargs)
        dissolved_default_units = self.return_default_units(sample, dissolved,**kwargs)
        return dissolved_default_units

    def check_calibration_range(self,sample,pressure,**kwargs):
        parameters = kwargs
        parameters['sample'] = sample
        parameters.update(dict(sample.get_composition(units='wtpt_oxides',normalization='none')))
        parameters['pressure'] = pressure
        if len(self.model.volatile_species) == 1:
            volspec = self.model.volatile_species[0]
            volconc = self.result
            parameters.update({volspec:volconc})
        else:
             parameters.update(self.result)

        calib_check = self.model.check_calibration_range(parameters)
        return calib_check

class calculate_equilibrium_fluid_comp(Calculate):
    """ Calculates the equilibrium fluid composition using a chosen model (default is MagmaSat).
    Using this interface will preprocess the sample, run the calculation, and then check
    the calibration ranges. All parameters required by the chosen model must be passed.

    Parameters
    ----------
    sample:    Sample class
        The rock composition as a Sample object.
    pressure:   float or None
        Total pressure in bars. If None, the saturation pressure will be used.
    model:  string or Model object
        Model to be used. If using one of the default models, this can be
        the string corresponding to the model in the default_models dict.
    silence_warnings     bool
        If set to True, no warnings will be raised automatically when calibration checks fail.
    preprocess_sample     bool
        If True (default), the sample will be preprocessed according to the preprocessing operations within
        the models. If you obtain unexpected results, try setting to False.

    Returns
    -------
    Calculate object
        Calculate object, access result by fetching the result property. Mole fractions
        of each volatile species, in order (CO2, then H2O, if using a mixed-fluid default
        model).
    """
    def calculate(self,sample,pressure=None,**kwargs):
        if type(pressure) == type(None):
            pressure = float(self.model.calculate_saturation_pressure(sample=sample,verbose=False,**kwargs))

        fluid_comp = self.model.calculate_equilibrium_fluid_comp(pressure=pressure,sample=sample,**kwargs)
        return fluid_comp

    def check_calibration_range(self,sample,pressure=None,**kwargs):
        if type(pressure) == type(None):
            pressure = self.model.calculate_saturation_pressure(sample=sample,**kwargs)
        parameters = kwargs
        parameters.update(dict(sample.get_composition(units='wtpt_oxides',normalization='none')))
        parameters['sample'] = sample
        parameters['pressure'] = pressure
        if len(self.model.volatile_species) == 1:
            volspec = self.model.volatile_species
            volconc = {volspec[0]:self.result}
            parameters.update(volconc)
        elif type(self.model.volatile_species) == list:
             parameters.update(self.result)

        calib_check = self.model.check_calibration_range(parameters)
        return calib_check


class calculate_isobars_and_isopleths(Calculate):
    """ Calculates isobars and isopleths using a chosen model (default is MagmaSat).
    Using this interface will preprocess the sample, run the calculation, and then check
    the calibration ranges. All parameters required by the chosen model must be passed.

    Parameters
    ----------
    sample:    Sample class
        The rock composition as a Sample object.
    pressure_list:   list
        List of all pressure values at which to calculate isobars, in bars.
    isopleth_list:   list
        OPTIONAL: Default value is None, in which case only isobars will be calculated. List of all
        fluid compositions in mole fraction (of the first species in self.volatile_species) at which
        to calcualte isopleths. Values can range from 0 to 1.
    points:     int
        The number of points in each isobar and isopleth. Default value is 101.
    model:  string or Model object
        Model to be used. If using one of the default models, this can be
        the string corresponding to the model in the default_models dict.
    silence_warnings     bool
        If set to True, no warnings will be raised automatically when calibration checks fail.
    preprocess_sample     bool
        If True (default), the sample will be preprocessed according to the preprocessing operations within
        the models. If you obtain unexpected results, try setting to False.

    Returns
    -------
    Calculate object
        Calculate object, access results by fetching the result property.
        If isopleth_list is not None, two objects will be returned, one with the isobars and the second with
        the isopleths. If return_dfs is True, two pandas DataFrames will be returned with column names
        'Pressure' or 'XH2O_fl', 'H2O_liq', and 'CO2_liq'. If return_dfs is False, two lists of numpy arrays
        will be returned. Each array is an individual isobar or isopleth, in the order passed via pressure_list
        or isopleth_list. The arrays are the concentrations of H2O and CO2 in the liquid, in the order of the
        species in self.volatile_species.
    """
    def calculate(self,sample,pressure_list,isopleth_list=[0,1],points=101,**kwargs):
        check = getattr(self.model, "calculate_isobars_and_isopleths", None)
        if callable(check):
            # samplenorm = sample.copy()
            # samplenorm = normalize_AdditionalVolatiles(samplenorm)
            isobars, isopleths = self.model.calculate_isobars_and_isopleths(sample=self.sample,pressure_list=pressure_list,isopleth_list=isopleth_list,points=points,**kwargs)
            return isobars, isopleths
        else:
            raise core.InputError("This model does not have a calculate_isobars_and_isopleths method built in, most likely because it is a pure fluid model.")

    def check_calibration_range(self,sample,pressure_list,**kwargs):
        parameters = kwargs
        parameters.update(dict(sample.get_composition(units='wtpt_oxides')))
        parameters['sample'] = sample
        s = ''
        s += self.model.check_calibration_range(parameters)
        parameters = {}
        if isinstance(pressure_list, list):
            pass
        else:
            pressure_list = [pressure_list]
        for pressure in pressure_list:
            parameters['pressure'] = pressure
            s += self.model.check_calibration_range(parameters,report_nonexistance=False)
        return s


class calculate_saturation_pressure(Calculate):
    """
    Calculates the pressure at which a fluid will be saturated, given the dissolved volatile
    concentrations. Using this interface will preprocess the sample, run the calculation, and then check
    the calibration ranges. All parameters required by the chosen model must be passed.

    Parameters
    ----------
    sample:    Sample class
        The rock composition as a Sample object.
    model:  string or Model object
        Model to be used. If using one of the default models, this can be
        the string corresponding to the model in the default_models dict.
    silence_warnings     bool
        If set to True, no warnings will be raised automatically when calibration checks fail.
    preprocess_sample     bool
        If True (default), the sample will be preprocessed according to the preprocessing operations within
        the models. If you obtain unexpected results, try setting to False.

    Returns
    -------
    Calculate object
        Calculate object, access results by fetching the result property.
        The saturation pressure in bars as a float.
    """
    def calculate(self,sample,**kwargs):
        satP = self.model.calculate_saturation_pressure(sample=sample,**kwargs)
        return satP

    def check_calibration_range(self,sample,**kwargs):
        parameters = kwargs
        if isinstance(self.result, dict): #handles cases where verbose=True
            parameters['pressure'] = next(iter(self.result.values()))
        else:
            parameters['pressure'] = self.result
        parameters.update(dict(sample.get_composition(units='wtpt_oxides')))
        parameters['sample'] = sample
        s = self.model.check_calibration_range(parameters)
        return s

class calculate_degassing_path(Calculate):
    """
    Calculates the dissolved volatiles in a progressively degassing sample.

    Parameters
    ----------
    sample:    Sample class
        The rock composition as a Sample object.
    pressure     string, float, int, list, or numpy array
        Defaults to 'saturation', the calculation will begin at the saturation pressure. If a number is passed
        as either a float or int, this will be the starting pressure. If a list of numpy array is passed, the
        pressure values in the list or array will define the degassing path, i.e. final_pressure and steps
        variables will be ignored. Units are bars.
    fractionate_vapor     float
        What proportion of vapor should be removed at each step. If 0.0 (default), the degassing path will
        correspond to closed-system degassing. If 1.0, the degassing path will correspond to open-system
        degassing.
    final_pressure         float
        The final pressure on the degassing path, in bars. Ignored if a list or numpy array is passed as the
        pressure variable. Default is 1 bar.
    steps     int
        The number of steps in the degassing path. Ignored if a list or numpy array are passed as the pressure
        variable.
    model:  string or Model object
        Model to be used. If using one of the default models, this can be
        the string corresponding to the model in the default_models dict.
    silence_warnings     bool
        If set to True, no warnings will be raised automatically when calibration checks fail.
    preprocess_sample     bool
        If True (default), the sample will be preprocessed according to the preprocessing operations within
        the models. If you obtain unexpected results, try setting to False.

    Returns
    -------
    Calculate object
        Calculate object, access results by fetching the result property.
        A DataFrame with columns 'Pressure', 'H2O_liq', 'CO2_liq',
        'H2O_fl', 'CO2_fl', and 'FluidProportion_wt', is returned. Dissolved volatiles are in wt%,
        the proportions of volatiles in the fluid are in mole fraction.
    """
    def calculate(self,sample,pressure='saturation',fractionate_vapor=0.0,
                  final_pressure=100.0,**kwargs):
        check = getattr(self.model, "calculate_degassing_path", None)
        if callable(check):
            data = self.model.calculate_degassing_path(sample=sample, pressure=pressure, fractionate_vapor=fractionate_vapor,
                                                       final_pressure=final_pressure, **kwargs)
            return data
        else:
            raise core.InputError("This model does not have a calculate_isobars_and_isopleths method built in, most likely because it is a pure fluid model.")

    def check_calibration_range(self,sample,**kwargs):
        parameters = kwargs
        parameters.update(dict(sample.get_composition()))
        parameters['sample'] = sample
        s = self.model.check_calibration_range(parameters)
        parameters = {}
        parameters['pressure'] = np.nanmax(self.result.Pressure_bars)
        # if istype(kwargs.get("pressure"), float) or istype(kwargs.get("pressure"), int):
        #     parameters['pressure'] = kwargs.get("pressure")
        s += self.model.check_calibration_range(parameters,report_nonexistance=False)
        return s



from copy import copy
import numpy as np

class CalibrationRange(object):
    """ The CalibrationRange object allows the range of allowable parameters to be specified and
    used in checking and reporting of the results.
    """
    def __init__(self, parameter_name, value, checkfunction=None, units='', model_name='',
                fail_msg='',fail_dict={}, pass_msg='', pass_dict={}, description_msg='', description_dict={}):
        self.parameter_name = parameter_name
        self.value = value
        self.checkfunction = checkfunction
        self.units = units
        self.model_name = model_name
        self.fail_msg = (copy(fail_msg), copy(fail_dict))
        self.pass_msg = (copy(pass_msg), copy(pass_dict))
        self.description_msg = (copy(description_msg), copy(description_dict))

    def check(self,parameters):
        """Method for checking whether parameters satisfy the calibration range."""
        if self.parameter_name in parameters:
            return self.checkfunction(self.value,parameters[self.parameter_name])
        else:
            return None

    def string(self,parameters,report_nonexistance=True):
        """Returns a string statement of the calibration check"""
        if type(parameters) == type(None):
            msgdict = self.description_msg[1]
            if type(self.value) == float or type(self.value) == int:
                msgdict['calib_val'] = self.value
            elif type(self.value) == list or type(self.value) == tuple or type(self.value) == np.ndarray:
                for i in range(len(self.value)):
                    msgdict['calib_val'+str(i)] = self.value[i]
            if 'param_name' not in msgdict:
                msgdict['param_name'] = self.parameter_name
            if 'units' not in msgdict:
                msgdict['units'] = self.units
            if 'model_name' not in msgdict:
                msgdict['model_name'] = self.model_name
            return self.description_msg[0].format(**msgdict)
        else:
            check = self.check(parameters)
            if check == True:
                msgdict = self.pass_msg[1]
                msgdict['param_val'] = parameters[self.parameter_name]
                if type(self.value) == float or type(self.value) == int:
                    msgdict['calib_val'] = self.value
                elif type(self.value) == list or type(self.value) == tuple or type(self.value) == np.ndarray:
                    for i in range(len(self.value)):
                        msgdict['calib_val'+str(i)] = self.value[i]
                if 'param_name' not in msgdict:
                    msgdict['param_name'] = self.parameter_name
                if 'units' not in msgdict:
                    msgdict['units'] = self.units
                if 'model_name' not in msgdict:
                    msgdict['model_name'] = self.model_name
                return self.pass_msg[0].format(**msgdict)
            elif check == False:
                msgdict = self.fail_msg[1]
                msgdict['param_val'] = parameters[self.parameter_name]
                if type(self.value) == float or type(self.value) == int:
                    msgdict['calib_val'] = self.value
                elif type(self.value) == list or type(self.value) == tuple or type(self.value) == np.ndarray:
                    for i in range(len(self.value)):
                        msgdict['calib_val'+str(i)] = self.value[i]
                if 'param_name' not in msgdict:
                    msgdict['param_name'] = self.parameter_name
                if 'units' not in msgdict:
                    msgdict['units'] = self.units
                if 'model_name' not in msgdict:
                    msgdict['model_name'] = self.model_name
                return self.fail_msg[0].format(**msgdict)
            else:
                if report_nonexistance == True:
                    return "A value for {} was not provided.".format(self.parameter_name)
                else:
                    return ''

# ------------- DEFAULT CALIBRATIONRANGE OBJECTS --------------- # FIND CALIBRATION RANGES

def crf_EqualTo(calibval,paramval):
    return calibval == paramval
crmsg_EqualTo_pass = "The {param_name} ({param_val:.1f} {units}) is equal to {calib_val:.1f} {units} as required by the calibration range of the {model_name} model. "
crmsg_EqualTo_fail ="{param_name} ({param_val:.1f} {units}) is outside the calibration range of the {model_name} model ({calib_val:.1f} {units}). "
crmsg_EqualTo_description = "The {model_name} model is calibrated for {param_name} equal to {calib_val:.1f} {units}. "

def crf_MixedFluidWarning(calibval,paramval):
    if type(paramval) == float or type(paramval) == int or len(paramval) == 1:
        return paramval == 0 or paramval == 1
    elif len(paramval) == 2:
        return list(paramval) == [0,1] or list(paramval) == [1,0]
    else:
        return True
crmsg_MixedFluidWarning_pass = "The pure fluid model is being used. "
crmsg_MixedFluidWarning_fail ="The {model_name} model should be used to model mixed fluids with caution. The mixed fluid model does not recreate the experimental observations. "
crmsg_MixedFluidWarning_description = "The {model_name} shoule be used to model mixed fluids with caution. "


def crf_GreaterThan(calibval,paramval):
    return paramval >= calibval
crmsg_GreaterThan_pass = "The {param_name} ({param_val:.1f} {units}) is greater than {calib_val:.1f} {units} as required by the calibration range of the {model_name} model. "
crmsg_GreaterThan_fail = "The {param_name} is outside the calibration range of {model_name} ({param_val:.1f}<{calib_val:.1f} {units}. "
crmsg_GreaterThan_description = "The {model_name} model is calibrated for {param_name} greater than {calib_val:.1f} {units}. "
# Warnings for specific models


def crf_LessThan(calibval,paramval):
    return paramval <= calibval
crmsg_LessThan_pass = "The {param_name} ({param_val:.1f} {units}) is less than {calib_val:.1f} {units} as required by the calibration range of the {model_name} model. "
crmsg_LessThan_fail =  "The {param_name} is outside the calibration range of {model_name} ({param_val:.1f}>{calib_val:.1f} {units}. "
crmsg_LessThan_description = "The {model_name} model is calibrated for {param_name} less than {calib_val:.1f} {units}. "
# Warnings for specific model

def crf_Between(calibval,paramval):
    return paramval >= calibval[0] and paramval <= calibval[1]
crmsg_Between_pass = "The {param_name} ({param_val:.1f} {units}) is between {calib_val0:.1f} and {calib_val1:.1f} {units} as required by the calibration range of the {model_name} model. "
crmsg_Between_fail = "{param_name} ({param_val:.1f} {units}) is outside the calibration range of the {model_name} model ({calib_val0:.1f}-{calib_val1:.1f} {units}). "
crmsg_Between_description = "The {model_name} model is calibrated for {param_name} between {calib_val0:.1f} and {calib_val1:.1f} {units}. "
# Different wording for compositional checks (loosing "the")
crmsg_BC_pass = "{param_name} ({param_val:.1f} {units}) is between {calib_val0:.1f} and {calib_val1:.1f} {units} as required by the calibration range of the {model_name} model. "
crmsg_BC_fail = "{param_name} ({param_val:.1f} {units}) is outside the calibration range ({calib_val0:.1f}-{calib_val1:.1f} {units}). "



from abc import abstractmethod
import numpy as np
import pandas as pd
import warnings as w
from scipy.optimize import root_scalar
from scipy.optimize import root
from copy import deepcopy

from VESIcal import activity_models
from VESIcal import core
from VESIcal import fugacity_models


class Model(object):
    """The model object implements a volatile solubility model. It is composed
    of the methods needed to evaluate :func:`VESIcal.calculate_dissolved_volatiles`,
    :func:`VESIcal.calculate_equilibrium_fluid_comp`, and :func:`calculate_saturation_pressure`. The
    fugacity and activity models for the volatiles species must be specified,
    defaulting to ideal.
    """

    def __init__(self):
        self.set_volatile_species(None)
        self.set_fugacity_model(fugacity_models.fugacity_idealgas())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_calibration_ranges([])
        self.set_solubility_dependence(False)

    def set_volatile_species(self,volatile_species):
        if type(volatile_species) == str:
            volatile_species = [volatile_species]
        elif type(volatile_species) != list:
            raise core.InputError("volatile_species must be a str or list.")
        self.volatile_species = volatile_species

    def set_fugacity_model(self,fugacity_model):
        self.fugacity_model = fugacity_model

    def set_activity_model(self,activity_model):
        self.activity_model = activity_model

    def set_calibration_ranges(self,calibration_ranges):
        self.calibration_ranges = calibration_ranges

    def set_solubility_dependence(self,solubility_dependence):
        self.solubility_dependence = solubility_dependence

    def get_calibration_values(self,variable_names):
        """ Returns the values stored as the calibration range for the given variable(s).
        However, for checks where there is a single value- i.e. cr_GreaterThan or crf_LessThan,
        the logical operation will remain a mystery until someone figures out an elegant way
        of communicating it.

        Parameters
        ----------
        variable_names     str or list
            The name(s) of the variables you want the calibration ranges for.

        Returns
        -------
        list
            A list of values or tuples for the calibration ranges in the order given.
        """

        # Check if the variable name is passed as a string, and if so put it in a list
        if type(variable_names) == str:
            variable_names = [variable_names]

        calibration_values = []

        for var in variable_names:
            found_var = False
            for cr in self.calibration_ranges:
                if found_var == False:
                    if cr.parameter_name == var:
                        found_var = True
                        calibration_values.append(cr.value)
            if found_var == False:
                calibration_values.append(np.nan)

        return calibration_values

    @abstractmethod
    def calculate_dissolved_volatiles(self,**kwargs):
        pass

    @abstractmethod
    def calculate_equilibrium_fluid_comp(self,**kwargs):
        pass

    @abstractmethod
    def calculate_saturation_pressure(self,**kwargs):
        pass

    # @abstractmethod
    # def preprocess_sample(self,**kwargs):
    #     pass

    def check_calibration_range(self,parameters,report_nonexistance=True):
        """ Checks whether the given parameters are within the ranges defined by the
        CalibrationRange objects for the model and its fugacity and activity models. An empty
        string will be returned if all parameters are within the calibration range. If a
        parameter is not within the calibration range, a description of the problem will be
        returned in the string.

        Parameters
        ----------
        parameters     dict
            Dictionary keys are the names of the parameters to be checked, e.g., pressure
            temperature, SiO2, etc. Values are the values of each parameter. A complete set
            need not be given.

        Returns
        -------
        str
            String description of any parameters falling outside of the calibration range.
        """
        s = ''
        for cr in self.calibration_ranges:
            if cr.check(parameters) == False:
                s += cr.string(parameters,report_nonexistance)
        for cr in self.fugacity_model.calibration_ranges:
            if cr.check(parameters) == False:
                s += cr.string(parameters,report_nonexistance)
        for cr in self.activity_model.calibration_ranges:
            if cr.check(parameters) == False:
                s += cr.string(parameters,report_nonexistance)
        return s

    def get_calibration_range(self):
        """ Returns a string describing the calibration ranges defined by the CalibrationRange
        objects for each model, and its associated fugacity and activity models.

        Returns
        -------
        str
            String description of the calibration range objects."""
        s = ''
        for cr in self.calibration_ranges:
            s += cr.string(None)
        for cr in self.fugacity_model.calibration_ranges:
            s += cr.string(None)
        for cr in self.activity_model.calibration_ranges:
            s += cr.string(None)
        return s



# ------------ PURE FLUID MODELS ------------------------------- #







# ------------ MIXED FLUID MODELS ------------------------------- #
class MixedFluid(Model):
    """
    Implements the generic framework for mixed fluid solubility. Any set of pure fluid solubility
    models may be specified.
    """
    def __init__(self,models):
        """
        Initializes the mixed fluid model.

        Parameters
        ----------
        models     dictionary
            Dictionary with names of volatile species as keys, and the model objects as values.
        """
        self.models = tuple(model for model in models.values())
        self.set_volatile_species(list(models.keys()))

    def get_calibration_values(self,variable_names):
        """ Placeholder method to prevent an error when this generic method is called for a MixedFluid
        model.

        Returns
        -------
        np.nan
        """
        return np.nan

    # def preprocess_sample(self,sample):
    #     """ Returns sample, unmodified.
    #
    #     Parameters
    #     ----------
    #     sample         pandas Series or dict
    #         Major element oxides in wt%.
    #
    #     Returns
    #     -------
    #     pandas Series or dict
    #         Major element oxides in wt%.
    #     """
    #     if type(sample) != dict and type(sample) != pd.core.series.Series:
    #         raise core.InputError("sample must be a dict or a pandas Series.")
    #     _sample = sample.copy()
    #     _sample = self.models[0].preprocess_sample(_sample)
    #     return _sample

    def calculate_dissolved_volatiles(self,pressure,X_fluid,returndict=False,**kwargs):
        """
        Calculates the dissolved volatile concentrations in wt%, using each model's
        calculate_dissolved_volatiles method. At present the volatile concentrations are
        not propagated through.

        Parameters
        ----------
        pressure     float
            The total pressure in bars.
        X_fluid     float, numpy.ndarry, dict, pandas Series
            The mole fraction of each species in the fluid. If the mixed fluid model
            contains only two species (e.g. CO2 and H2O), the value of the first species in
            self.volatile_species may be passed on its own as a float.
        returndict         bool
            If True, the results will be returned in a dict, otherwise they will be returned
            as a tuple.

        Returns
        -------
        tuple
            Dissolved volatile concentrations of each species in the model, in the order set
            by self.volatile_species.
        """
        if (type(X_fluid) == float or type(X_fluid) == int) and len(self.volatile_species) == 2:
            X_fluid = (X_fluid,1-X_fluid)
        elif len(X_fluid) != len(self.volatile_species):
            raise core.InputError("X_fluid must have the same length as the number of volatile species\
             in the MixedFluids Model class, or it may have length 1 if two species are present\
             in the MixedFluids Model class.")

        if np.sum(X_fluid) != 1.0:
            raise core.InputError("X_fluid must sum to 1.0")
        if any(val<0 for val in X_fluid) or any(val>1 for val in X_fluid):
            raise core.InputError("Each mole fraction in X_fluid must have a value between 0 and 1.")

        if type(X_fluid) == dict or type(X_fluid) == pd.core.series.Series:
            X_fluid = tuple(X_fluid[species] for species in self.volatile_species)

        # If the models don't depend on the concentration of volatiles, themselves.
        if all(model.solubility_dependence == False for model in self.models):
            result = tuple(model.calculate_dissolved_volatiles(pressure=pressure,X_fluid=Xi,**kwargs) for model, Xi in zip(self.models,X_fluid))
        # If one of the models depends on the other volatile concentration
        elif len(self.models) == 2 and self.models[0].solubility_dependence == False and 'sample' in kwargs:
            result0 = self.models[0].calculate_dissolved_volatiles(pressure=pressure,X_fluid=X_fluid[0],**kwargs)
            samplecopy = kwargs['sample'].change_composition({self.volatile_species[0]:result0},inplace=False)
            kwargs['sample'] = samplecopy
            result1 = self.models[1].calculate_dissolved_volatiles(pressure=pressure,X_fluid=X_fluid[1],**kwargs)
            result = (result0,result1)
        elif len(self.models) == 2 and self.models[1].solubility_dependence == False and 'sample' in kwargs:
            result1 = self.models[1].calculate_dissolved_volatiles(pressure=pressure,X_fluid=X_fluid[1],**kwargs)
            samplecopy = kwargs['sample'].change_composition({self.volatile_species[1]:result1}, inplace=False)
            kwargs['sample'] = samplecopy
            result0 = self.models[0].calculate_dissolved_volatiles(pressure=pressure,X_fluid=X_fluid[0],**kwargs)
            result = (result0,result1)
        else:
            raise core.InputError("The solubility dependence of the models is not currently supported by the MixedFluid model.")

        if returndict == True:
            resultsdict = {}
            for i,v in zip(range(len(self.volatile_species)),self.volatile_species):
                resultsdict.update({v+'_liq':result[i]})
            return resultsdict
        else:
            return result


    def calculate_equilibrium_fluid_comp(self,pressure,sample,return_dict=True,**kwargs):
        """ Calculates the composition of the fluid in equilibrium with the dissolved volatile
        concentrations passed. If a fluid phase is undersaturated at the chosen pressure (0,0) will
        be returned. Note, this currently assumes the given H2O and CO2 concentrations are
        the system total, not the total dissolved. If one of the volatile species has a zero or
        negative concentration, the pure fluid model for the other volatile species will be used.

        Parameters
        ----------
        pressure     float
            The total pressure in bars.
        sample    Sample class
            Magma major element composition.
        return_dict     bool
            Set the return type, if true a dict will be returned, if False two floats will be
            returned. Default is True.

        Returns
        -------
        dict or floats
            Mole fractions of the volatile species in the fluid, in the order given by
            self.volatile_species if floats.
        """
        if len(self.volatile_species) != 2:
            raise core.InputError("Currently equilibrium fluid compositions can only be calculated when\
            two volatile species are present.")

        dissolved_at_0bar = [self.models[0].calculate_dissolved_volatiles(sample=sample,pressure=0.0,**kwargs),
                             self.models[1].calculate_dissolved_volatiles(sample=sample,pressure=0.0,**kwargs)]

        if sample.get_composition(self.volatile_species[0]) <= 0.0 or sample.get_composition(self.volatile_species[0]) <= dissolved_at_0bar[0]:
            Xv0 = 0.0
            Xv1 = self.models[1].calculate_equilibrium_fluid_comp(pressure=pressure,sample=sample,**kwargs)
        elif sample.get_composition(self.volatile_species[1]) <= 0.0 or sample.get_composition(self.volatile_species[1]) <= dissolved_at_0bar[1]:
            Xv1 = 0.0
            Xv0 = self.models[0].calculate_equilibrium_fluid_comp(pressure=pressure,sample=sample,**kwargs)
        else:
            satP = self.calculate_saturation_pressure(sample,**kwargs)

            if satP < pressure:
                if return_dict == True:
                    return {self.volatile_species[0]:0,self.volatile_species[1]:0}
                else:
                    return (0,0)

            molfracs = sample.get_composition(units='mol_oxides')
            (Xt0, Xt1) = (molfracs[self.volatile_species[0]],molfracs[self.volatile_species[1]])

            try:
                Xv0 = root_scalar(self.root_for_fluid_comp,bracket=[1e-15,1-1e-15],args=(pressure,Xt0,Xt1,sample,kwargs)).root
                Xv1 = 1 - Xv0
            except:
                try:
                    Xv0 = root_scalar(self.root_for_fluid_comp,x0=0.5,x1=0.1,args=(pressure,Xt0,Xt1,sample,kwargs)).root
                    Xv1 = 1 - Xv0
                except:
                    raise core.SaturationError("Equilibrium fluid not found. Likely an issue with the numerical solver.")

        if return_dict == True:
            return {self.volatile_species[0]:Xv0,self.volatile_species[1]:Xv1}
        else:
            return Xv0, Xv1


    def calculate_saturation_pressure(self,sample,**kwargs):
        """
        Calculates the pressure at which a fluid will be saturated, given the dissolved volatile
        concentrations. If one of the volatile species has a zero or negative concentration the
        pure fluid model for the other species will be used. If one of the volatile species has a
        concentration lower than the concentration dissolved at 0 bar, the pure fluid model for the
        other species will be used.

        Parameters
        ----------
        sample     Sample class
            Magma major element composition (including volatiles).

        Returns
        -------
        float
            The saturation pressure in bars.
        """
        dissolved_at_0bar = [self.models[0].calculate_dissolved_volatiles(sample=sample,pressure=0.0,**kwargs),
                             self.models[1].calculate_dissolved_volatiles(sample=sample,pressure=0.0,**kwargs)]

        if sample.get_composition(self.volatile_species[0]) <= 0.0 or sample.get_composition(self.volatile_species[0]) <= dissolved_at_0bar[0]:
            satP = self.models[1].calculate_saturation_pressure(sample=sample,**kwargs)
        elif sample.get_composition(self.volatile_species[1]) <= 0.0 or sample.get_composition(self.volatile_species[1]) <= dissolved_at_0bar[1]:
            satP = self.models[0].calculate_saturation_pressure(sample=sample,**kwargs)
        else:
            volatile_concs = np.array(tuple(sample.get_composition(species) for species in self.volatile_species))

            x0 = 0
            for model in self.models:
                xx0 = model.calculate_saturation_pressure(sample=sample,**kwargs)
                if np.isnan(xx0) == False:
                    x0 += xx0

            try:
                satP = root(self.root_saturation_pressure,x0=[x0,0.5],args=(volatile_concs,sample,kwargs)).x[0]
            except:
                w.warn("Saturation pressure not found.",RuntimeWarning,stacklevel=2)
                satP = np.nan

        return satP

    def calculate_isobars_and_isopleths(self,pressure_list,isopleth_list=[0,1],points=51,
                                        return_dfs=True,extend_to_zero=True,**kwargs):
        """
        Calculates isobars and isopleths. Isobars can be calculated for any number of pressures. Variables
        required by each of the pure fluid models must be passed, e.g. sample, temperature, etc.

        Parameters
        ----------
        pressure_list     list or float
            List of all pressure values at which to calculate isobars, in bars.
        isopleth_list     list
            Default value is None, in which case only isobars will be calculated. List of all
            fluid compositions in mole fraction (of the first species in self.volatile_species) at which
            to calcualte isopleths. Values can range from 0 to 1.
        points     int
            The number of points in each isobar and isopleth. Default value is 101.
        return_dfs     bool
            If True, the results will be returned as two pandas DataFrames, as produced by the MagmaSat
            method. If False the results will be returned as lists of numpy arrays.

        Returns
        -------
        pandas DataFrame object(s) or list(s)
            If isopleth_list is not None, two objects will be returned, one with the isobars and the second with
            the isopleths. If return_dfs is True, two pandas DataFrames will be returned with column names
            'Pressure' or 'XH2O_fl', 'H2O_liq', and 'CO2_liq'. If return_dfs is False, two lists of numpy arrays
            will be returned. Each array is an individual isobar or isopleth, in the order passed via pressure_list
            or isopleth_list. The arrays are the concentrations of H2O and CO2 in the liquid, in the order of the
            species in self.volatile_species.

        """
        if len(self.volatile_species) != 2 or 'H2O' not in self.volatile_species or 'CO2' not in self.volatile_species:
            raise core.InputError("calculate_isobars_and_isopleths may only be used with a H2O-CO2 fluid.")

        H2O_id = self.volatile_species.index('H2O')
        CO2_id = self.volatile_species.index('CO2')

        if isinstance(pressure_list, list):
            pass
        elif isinstance(pressure_list, int) or isinstance(pressure_list, float):
            pressure_list = [pressure_list]
        else:
            raise core.InputError("pressure_list must be a single float (1000.0), int (1000), or list of those [1000, 2000.0, 3000].")

        has_isopleths = True
        if isopleth_list is None:
            has_isopleths = False


        isobars_df = pd.DataFrame(columns=['Pressure','H2O_liq','CO2_liq'])
        isobars = []
        for pressure in pressure_list:
            dissolved = np.zeros([2,points])
            Xv0 = np.linspace(0.0,1.0,points)
            for i in range(points):
                dissolved[:,i] = self.calculate_dissolved_volatiles(pressure=pressure,X_fluid=(Xv0[i],1-Xv0[i]),**kwargs)
                isobars_df = isobars_df.append({'Pressure':pressure,'H2O_liq':dissolved[H2O_id,i],'CO2_liq':dissolved[CO2_id,i]},ignore_index=True)
            isobars.append(dissolved)


        if has_isopleths == True:
            isopleths_df = pd.DataFrame(columns=['XH2O_fl','H2O_liq','CO2_liq'])
            isopleths = []
            for isopleth in isopleth_list:
                dissolved = np.zeros([2,points])
                pmin = np.nanmin(pressure_list)
                pmax = np.nanmax(pressure_list)
                if pmin == pmax:
                    pmin = 0.0
                pressure = np.linspace(pmin,pmax,points)
                for i in range(points):
                    dissolved[:,i] = self.calculate_dissolved_volatiles(pressure=pressure[i],X_fluid=(isopleth,1-isopleth),**kwargs)
                    isopleths_df = isopleths_df.append({'XH2O_fl':[isopleth,1-isopleth][H2O_id],'H2O_liq':dissolved[H2O_id,i],'CO2_liq':dissolved[CO2_id,i]},ignore_index=True)
                isopleths.append(dissolved)

        if return_dfs == True:
            if has_isopleths == True:
                return (isobars_df, isopleths_df)
            else:
                return isobars_df
        else:
            if has_isopleths == True:
                return (isobars, isopleths)
            else:
                return isobars


    def calculate_degassing_path(self,sample,pressure='saturation',fractionate_vapor=0.0,final_pressure=100.0,
                                    steps=101,return_dfs=True,round_to_zero=True,**kwargs):
        """
        Calculates the dissolved volatiles in a progressively degassing sample.

        Parameters
        ----------
        sample     Sample class
            Magma major element composition (including volatiles).
        pressure     string, float, int, list, or numpy array
            Defaults to 'saturation', the calculation will begin at the saturation pressure. If a number is passed
            as either a float or int, this will be the starting pressure. If a list of numpy array is passed, the
            pressure values in the list or array will define the degassing path, i.e. final_pressure and steps
            variables will be ignored. Units are bars.
        fractionate_vapor     float
            What proportion of vapor should be removed at each step. If 0.0 (default), the degassing path will
            correspond to closed-system degassing. If 1.0, the degassing path will correspond to open-system
            degassing.
        final_pressure         float
            The final pressure on the degassing path, in bars. Ignored if a list or numpy array is passed as the
            pressure variable. Default is 1 bar.
        steps     int
            The number of steps in the degassing path. Ignored if a list or numpy array are passed as the pressure
            variable.
        return_dfs     bool
            If True, the results will be returned in a pandas DataFrame, if False, two numpy arrays will be returned.
        round_to_zero   bool
            If True, the first entry of FluidProportion_wt will be rounded to zero, rather than being a value
            within numerical error of zero. Default is True.

        Returns
        -------
        pandas DataFrame or numpy arrays
            If return_dfs is True (default), a DataFrame with columns 'Pressure_bars', 'H2O_liq', 'CO2_liq',
            'H2O_fl', 'CO2_fl', and 'FluidProportion_wt', is returned. Dissolved volatiles are in wt%,
            the proportions of volatiles in the fluid are in mole fraction. Otherwise a numpy array containing
            the dissolved volatile concentrations, and a numpy array containing the mole fractions of
            volatiles in the fluid is returned. The columns are in the order of the volatiles in
            self.volatile_species.

        """

        # Create a copy of the sample so that initial volatile concentrations are not overwritten.
        sample = deepcopy(sample)

        # Its imperative that normalization doesn't change the volatile concentrations throughout
        # the calculation.
        if sample.default_normalization not in ['fixedvolatiles','none']:
            sample.set_default_normalization('fixedvolatiles')
            w.warn('Sample normalization changed to fixedvolatiles.')

        wtptoxides = sample.get_composition(units='wtpt_oxides')
        wtm0s, wtm1s = (wtptoxides[self.volatile_species[0]],wtptoxides[self.volatile_species[1]])

        if pressure == 'saturation':
            p0 = self.calculate_saturation_pressure(sample,**kwargs)
            pressures = np.linspace(p0,final_pressure,steps)
        elif type(pressure) == float or type(pressure) == int:
            pressures = np.linspace(pressure,final_pressure,steps)
        elif type(pressure) == list or type(pressure) == np.ndarray:
            pressures = pressure

        Xv = np.zeros([2,len(pressures)])
        wtm = np.zeros([2,len(pressures)])

        for i in range(len(pressures)):
            try:
                wtptoxides = sample.get_composition(units='wtpt_oxides')
                X_fluid = self.calculate_equilibrium_fluid_comp(pressure=pressures[i],sample=sample,return_dict=False,**kwargs)
                Xv[:,i] = X_fluid
                if X_fluid == (0,0):
                    wtm[:,i] = (wtptoxides[self.volatile_species[0]],wtptoxides[self.volatile_species[1]])
                else:
                    if X_fluid[0] == 0:
                        wtm[0,i] = wtptoxides[self.volatile_species[0]]
                        wtm[1,i] = self.calculate_dissolved_volatiles(pressure=pressures[i],sample=sample,X_fluid=X_fluid,**kwargs)[1]
                    elif X_fluid[1] == 0:
                        wtm[1,i] = wtptoxides[self.volatile_species[1]]
                        wtm[0,i] = self.calculate_dissolved_volatiles(pressure=pressures[i],sample=sample,X_fluid=X_fluid,**kwargs)[0]
                    else:
                        wtm[:,i] = self.calculate_dissolved_volatiles(pressure=pressures[i],sample=sample,X_fluid=X_fluid,**kwargs)


                    sample.change_composition({self.volatile_species[0]: wtm[0,i] + (1-fractionate_vapor)*(wtm0s-wtm[0,i]),
                                               self.volatile_species[1]: wtm[1,i] + (1-fractionate_vapor)*(wtm1s-wtm[1,i])})
            except:
                Xv[:,i] = [np.nan]*np.shape(Xv)[0]
                wtm[:,i] = wtm[:,i-1]

        if return_dfs == True:
            exsolved_degassing_df = pd.DataFrame()
            exsolved_degassing_df['Pressure_bars'] = pressures
            exsolved_degassing_df['H2O_liq'] = wtm[self.volatile_species.index('H2O'),:]
            exsolved_degassing_df['CO2_liq'] = wtm[self.volatile_species.index('CO2'),:]
            exsolved_degassing_df['H2O_fl'] = Xv[self.volatile_species.index('H2O'),:]
            exsolved_degassing_df['CO2_fl'] = Xv[self.volatile_species.index('CO2'),:]
            exsolved_degassing_df['FluidProportion_wt'] = (wtm0s+wtm1s)-exsolved_degassing_df['H2O_liq']-exsolved_degassing_df['CO2_liq']

            if round_to_zero == True and np.round(exsolved_degassing_df.loc[0,'FluidProportion_wt'],2)==0:
                exsolved_degassing_df.loc[0,'FluidProportion_wt'] = 0.0

            return exsolved_degassing_df

        else:
            return (wtm, Xv)


    def root_saturation_pressure(self,x,volatile_concs,sample,kwargs):
        """ Function called by scipy.root when finding the saturation pressure using
        calculate_saturation_pressure.

        Parameters
        ----------
        x     numpy array
            The guessed value for the root. x[0] is the pressure (in bars) and x[1] is the
            mole fraction of the first volatile in self.volatile_species.
        volatile_concs     numpy array
            The dissolved volatile concentrations, in the same order as self.volatile_species.
        sample:     Sample class
            Magma major element composition (including volatiles).
        kwargs     dictionary
            Dictionary of keyword arguments, which may be required by the pure-fluid models.

        Returns
        -------
        numpy array
            The difference in the dissolved volatile concentrations, and those predicted with the
            pressure and fluid composition specified by x.
        """
        if x[1] < 0:
            x[1] = 0
        elif x[1] > 1:
            x[1] = 1
        if x[0] <= 0:
            x[0] = 1e-15
        misfit = np.array(self.calculate_dissolved_volatiles(pressure=x[0],X_fluid=(x[1],1-x[1]),sample=sample,**kwargs)) - volatile_concs
        return misfit


    def root_for_fluid_comp(self,Xv0,pressure,Xt0,Xt1,sample,kwargs):
        """ Function called by scipy.root_scalar when calculating the composition of equilibrium fluid
        in the calculate_equilibrium_fluid_comp method.

        Parameters
        ----------
        Xv0     float
            The guessed mole fraction of the first volatile species in self.volatile_species.
        pressure     float
            The total pressure in bars.
        Xt0     float
            The total mole fraction of the first volatile species in self.volatile_species.
        Xt1        float
            The total mole fraction of the second volatile species in self.volatile_species.
        sample     Sample class
            Magma major element composition.
        kwargs     dictionary
            A dictionary of keyword arguments that may be required by the pure fluid models.

        Returns
        -------
        float
            The differene in the LHS and RHS of the mass balance equation. Eq X in manuscript.
        """

        wtt0 = sample.get_composition(self.volatile_species[0])
        wtt1 = sample.get_composition(self.volatile_species[1])

        wtm0, wtm1 = self.calculate_dissolved_volatiles(pressure=pressure,X_fluid=(Xv0,1-Xv0),sample=sample,**kwargs)

        Xm0 = Xt0/wtt0*wtm0
        Xm1 = Xt1/wtt1*wtm1

        if self.volatile_species[0] == 'CO2' and Xv0 != Xm0:
            f = (Xt0-Xm0)/(Xv0-Xm0)
            return (1-f)*Xm1 + f*(1-Xv0) - Xt1
        else:
            f = (Xt1-Xm1)/((1-Xv0)-Xm1)
            return (1-f)*Xm0 + f*Xv0 - Xt0

    def check_calibration_range(self,parameters,report_nonexistance=True):
        """ Checks whether the given parameters are within the ranges defined by the
        CalibrationRange objects for each model and its fugacity and activity models. An empty
        string will be returned if all parameters are within the calibration range. If a
        parameter is not within the calibration range, a description of the problem will be
        returned in the string.

        Parameters
        ----------
        parameters     dict
            Dictionary keys are the names of the parameters to be checked, e.g., pressure
            temperature, SiO2, etc. Values are the values of each parameter. A complete set
            need not be given.

        Returns
        -------
        str
            String description of any parameters falling outside of the calibration range.
        """
        s = ''
        for model in self.models:
            for cr in model.calibration_ranges:
                if cr.check(parameters) == False:
                    s += cr.string(parameters,report_nonexistance)
            for cr in model.fugacity_model.calibration_ranges:
                if cr.check(parameters) == False:
                    s += cr.string(parameters,report_nonexistance)
            for cr in model.activity_model.calibration_ranges:
                if cr.check(parameters) == False:
                    s += cr.string(parameters,report_nonexistance)
        return s

    def get_calibration_range(self):
        """ Returns a string describing the calibration ranges defined by the CalibrationRange
        objects for each model, and its associated fugacity and activity models.

        Returns
        -------
        str
            String description of the calibration range objects."""
        s = ''
        for model in self.models:
            for cr in model.calibration_ranges:
                s += cr.string(None)
            for cr in model.fugacity_model.calibration_ranges:
                s += cr.string(None)
            for cr in model.activity_model.calibration_ranges:
                s += cr.string(None)
        return s



from VESIcal import activity_models
from VESIcal import calibration_checks
from VESIcal import core
from VESIcal import fugacity_models
from VESIcal import model_classes
from VESIcal import sample_class

import numpy as np
import warnings as w
from scipy.optimize import root_scalar


class carbon(model_classes.Model):
    """ Implementation of the Shishkina et al. (2014) carbon solubility model, as a Model class.
    """
    def __init__(self):
        self.set_volatile_species(['CO2'])
        self.set_fugacity_model(fugacity_models.fugacity_idealgas())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_solubility_dependence(False)
        self.set_calibration_ranges([calibration_checks.CalibrationRange('pressure',[500.0,5000.0],calibration_checks.crf_Between,'bar','Shishkina et al. carbon',
                                                      fail_msg=calibration_checks.crmsg_Between_fail,
                                                      pass_msg=calibration_checks.crmsg_Between_pass,
                                                      description_msg=calibration_checks.crmsg_Between_description),
                                     calibration_checks.CalibrationRange('temperature',[1200.0,1300.0],calibration_checks.crf_Between,'oC','Shishkina et al. carbon',
                                                       fail_msg=calibration_checks.crmsg_Between_fail,
                                                      pass_msg=calibration_checks.crmsg_Between_pass,
                                                      description_msg=calibration_checks.crmsg_Between_description),
                                     calibration_checks.CalibrationRange('SiO2',[40,57],calibration_checks.crf_Between,'wt%','Shishkina et al. carbon',
                                                       fail_msg=crmsg_BC_fail_C1,
                                                      pass_msg=calibration_checks.crmsg_Between_pass,
                                                      description_msg=calibration_checks.crmsg_Between_description)])

    def PiStar(self,sample):
        """Shishkina et al. (2014) Eq (11)

        Calculates the Pi* parameter for use in calculating CO2 solubility.

        Parameters
        ----------
        sample:        Sample class
            The magma composition stored in a Sample class.

        Returns
        -------
        float
            The value of the Pi* compositional parameter.
        """

        _mols = sample.get_composition(units='mol_cations')

        if all(cation in _mols for cation in ['Ca','K','Na','Mg','Fe','Si','Al']) == False:
            raise core.InputError("To calculate PiStar, values for CaO, K2O, Na2O, MgO, FeO, SiO2, and Al2O3\
                                must be provided in sample.")

        _pi = (_mols['Ca'] + 0.8*_mols['K'] + 0.7*_mols['Na'] + 0.4*_mols['Mg'] + 0.4*_mols['Fe'])/\
                (_mols['Si']+_mols['Al'])

        return _pi

    def calculate_dissolved_volatiles(self,pressure,sample,X_fluid=1,**kwargs):
        """ Calculates the dissolved CO2 concentration in wt%, using equation (13) of Shishkina et al. (2014).

        Parameters
        ----------
        pressure:    float
            (Total) pressure in bars.
        sample:        Sample class
            Magma composition.
        X_fluid:    float
            The mol-fraction of the fluid that is CO2. Default is 1, i.e. a pure CO2 fluid.

        Returns
        -------
        float
            The dissolved CO2 concentration in wt%.
        """

        if X_fluid < 0 or X_fluid > 1:
            raise core.InputError("X_fluid must have a value between 0 and 1.")
        if pressure < 0:
            raise core.InputError("pressure must be a positive value.")

        PiStar = self.PiStar(sample)
        fugacity = self.fugacity_model.fugacity(pressure=pressure,X_fluid=X_fluid,**kwargs)

        A = 1.150
        B = 6.71
        C= -1.345

        if fugacity == 0:
            return 0
        else:
            return np.exp(A*np.log(fugacity/10)+B*PiStar+C)/1e4


    def calculate_equilibrium_fluid_comp(self,pressure,sample,**kwargs):
        """ Returns 1.0 if a pure CO2 fluid is saturated.
        Returns 0.0 if a pure CO2 fluid is undersaturated.

        Parameters
        ----------
        pressure     float
            The total pressure of the system in bars.
        sample         Sample class
            Magma major element composition.

        Returns
        -------
        float
            1.0 if CO2-fluid saturated, 0.0 otherwise.
        """
        if self.calculate_saturation_pressure(sample=sample,**kwargs) < pressure:
            return 0.0
        else:
            return 1.0

    def calculate_saturation_pressure(self,sample,**kwargs):
        """ Calculates the pressure at which a pure CO2 fluid is saturated, for the given
        sample composition and CO2 concentration. Calls the scipy.root_scalar routine, which makes
        repeated calls to the calculate_dissolved_volatiles method.

        Parameters
        ----------
        sample         Sample class
            Magma major element composition.

        Returns
        -------
        float
            Saturation pressure in bar
        """

        if sample.check_oxide('CO2') == False:
            raise core.InputError("sample must contain CO2.")
        if sample.get_composition('CO2') < 0:
            raise core.InputError("CO2 concentration must be greater than 0 wt%.")

        try:
            satP = root_scalar(self.root_saturation_pressure,bracket=[1e-15,1e5],args=(sample,kwargs)).root
        except:
            w.warn("Saturation pressure not found.",RuntimeWarning,stacklevel=2)
            satP = np.nan
        return satP

    def root_saturation_pressure(self,pressure,sample,kwargs):
        """ Function called by scipy.root_scalar when finding the saturation pressure using
        calculate_saturation_pressure.

        Parameters
        ----------
        pressure     float
            Pressure guess in bars
        sample         Sample class
            Magma major element composition.
        kwargs         dictionary
            Additional keyword arguments supplied to calculate_saturation_pressure. Might be required for
            the fugacity or activity models.

        Returns
        -------
        float
            The differece between the dissolved CO2 at the pressure guessed, and the CO2 concentration
            passed in the sample variable.
        """
        return self.calculate_dissolved_volatiles(pressure=pressure,sample=sample,**kwargs)-\
            sample.get_composition('CO2')



class water(model_classes.Model):
    """ Implementation of the Shishkina et al. (2014) H2O solubility model as a Model class.
    """
    def __init__(self):
        self.set_volatile_species(['H2O'])
        self.set_fugacity_model(fugacity_models.fugacity_idealgas())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_solubility_dependence(False)
        self.set_calibration_ranges([calibration_checks.CalibrationRange('pressure',[500.0,5000.0],calibration_checks.crf_Between,'bar','Shishkina et al. water',
                                                      fail_msg=calibration_checks.crmsg_Between_fail,
                                                      pass_msg=calibration_checks.crmsg_Between_pass,
                                                      description_msg=calibration_checks.crmsg_Between_description),
                                    calibration_checks.CalibrationRange('temperature',[1050,1400],calibration_checks.crf_Between,'oC','Shishkina et al. water',
                                                       fail_msg=crmsg_BC_fail_T1,
                                                      pass_msg=calibration_checks.crmsg_Between_pass,
                                                       description_msg=calibration_checks.crmsg_Between_description),
                                    calibration_checks.CalibrationRange('SiO2',65,calibration_checks.crf_LessThan,'wt%','Shishkina et al. water',
                                                       fail_msg=crmsg_LessThan_fail_WaterSi,
                                                      pass_msg=calibration_checks.crmsg_LessThan_pass,
                                                       description_msg=calibration_checks.crmsg_LessThan_description),
                                    calibration_checks.CalibrationRange('SiO2',40,calibration_checks.crf_GreaterThan,'wt%','Shishkina et al. water',
                                                       fail_msg=crmsg_GreaterThan_fail_WaterSi,
                                                      pass_msg=calibration_checks.crmsg_GreaterThan_pass,
                                                       description_msg=calibration_checks.crmsg_GreaterThan_description)])

    def calculate_dissolved_volatiles(self,pressure,sample,X_fluid=1.0,**kwargs):
        """Calculates the dissolved H2O concentration using Eqn (9) of Shishkina et al. (2014).

        Parameters
        ----------
        pressure     float
            Total pressure in bars
        sample         Sample class
            Magma major element composition.
        X_fluid     float
            The mol fraction of H2O in the fluid

        Returns
        -------
        float
            The H2O concentration in wt%
        """
        if isinstance(sample,sample_class.Sample) == False:
            raise core.InputError("Sample must be an instance of the Sample class.")

        if sample.check_oxide('Na2O') == False or sample.check_oxide('K2O') == False:
            raise core.InputError("Na2O and K2O must be present in sample.")

        if pressure < 0:
            raise core.InputError("Pressure must be positive.")

        _mols = sample.get_composition(units='mol_cations')
        _mol_volatiles = 0
        if 'H' in _mols:
            _mol_volatiles += _mols['H']
        if 'C' in _mols:
            _mol_volatiles += _mols['C']

        total_alkalis = (_mols['Na'] + _mols['K'])/(1-_mol_volatiles)

        fugacity = self.fugacity_model.fugacity(pressure,X_fluid=X_fluid,**kwargs)

        a = 3.36e-7 * (fugacity/10)**3 - 2.33e-4*(fugacity/10)**2 + 0.0711*(fugacity/10) - 1.1309
        b = -1.2e-5*(fugacity/10)**2 + 0.0196*(fugacity/10)+1.1297

        return a*total_alkalis + b


    def calculate_equilibrium_fluid_comp(self,pressure,sample,**kwargs):
        """ Returns 1.0 if a pure H2O fluid is saturated.
        Returns 0.0 if a pure H2O fluid is undersaturated.

        Parameters
        ----------
        pressure     float
            The total pressure of the system in bars.
        sample       Sample class
            Magma major element composition (including H2O).

        Returns
        -------
        float
            1.0 if H2O-fluid saturated, 0.0 otherwise.
        """
        if self.calculate_saturation_pressure(sample=sample,**kwargs) < pressure:
            return 0.0
        else:
            return 1.0

    def calculate_saturation_pressure(self,sample,**kwargs):
        """ Calculates the pressure at which a pure H2O fluid is saturated, for the given
        sample composition and H2O concentration. Calls the scipy.root_scalar routine, which makes
        repeated calls to the calculate_dissolved_volatiles method.

        Parameters
        ----------
        sample         Sample class
            Magma major element composition (including H2O).

        Returns
        -------
        float
            Saturation pressure in bar
        """
        if sample.check_oxide('H2O') == False:
            raise core.InputError("sample must contain H2O")
        if sample.get_composition('H2O') < 0:
            raise core.InputError("H2O concentration must be greater than 0 wt%.")

        if sample.get_composition('H2O') < self.calculate_dissolved_volatiles(sample=sample,pressure=0,**kwargs):
            return np.nan

        try:
            satP = root_scalar(self.root_saturation_pressure,bracket=[1e-15,1e5],args=(sample,kwargs)).root
        except:
            w.warn("Saturation pressure not found.",RuntimeWarning,stacklevel=2)
            satP = np.nan
        return satP

    def root_saturation_pressure(self,pressure,sample,kwargs):
        """ Function called by scipy.root_scalar when finding the saturation pressure using
        calculate_saturation_pressure.

        Parameters
        ----------
        pressure     float
            Pressure guess in bars
        sample         Sample class
            Magma major element composition (including H2O).
        kwargs         dictionary
            Additional keyword arguments supplied to calculate_saturation_pressure. Might be required for
            the fugacity or activity models.

        Returns
        -------
        float
            The differece between the dissolved H2O at the pressure guessed, and the H2O concentration
            passed in the sample variable.
        """
        return self.calculate_dissolved_volatiles(pressure=pressure,sample=sample,**kwargs)-\
            sample.get_composition('H2O')


crmsg_BC_fail_T1 = "{param_name} ({param_val:.1f} {units}) is outside the calibration range of {model_name} \
({calib_val0:.1f}-{calib_val1:.1f} {units}). Note, the authors recomend that this model is optimally calibrated between 1150-1250C. "
crmsg_BC_fail_C1="{param_name} ({param_val:.1f} {units}) is outside the calibration range of {model_name} (calculated\
 from the max and minimum concentrations in the calibration dataset +-5%; {calib_val0:.1f}-{calib_val1:.1f} {units})."
crmsg_GreaterThan_fail_WaterSi = "{param_name}{units} exceeds the lower calibration limit of {model_name} ({calib_val:.1f} \
{units}, based on the minimum concentration minus 5% in the calibration dataset). " # This check warns users if SiO2<40 in the water model
crmsg_LessThan_fail_WaterSi = "{param_name} ({param_val:.1f} {units}) exceeds the upper limit of {calib_val:.1f} {units} \
suggested by Shishkina et al. for their H2O model. " # This check warns users if SiO2>65 wt% , the upper limit suggested by Shishkina et al. (2014)



mixed = model_classes.MixedFluid({'H2O':water(),'CO2':carbon()})

mixed.models[0].calibration_ranges.append(calibration_checks.CalibrationRange('X_fluid',0.0,calibration_checks.crf_MixedFluidWarning,'','Shishkina et al. (2014)',
                                                                                 fail_msg=calibration_checks.crmsg_MixedFluidWarning_fail,
                                                                                pass_msg=calibration_checks.crmsg_MixedFluidWarning_pass,
                                                                                description_msg=calibration_checks.crmsg_MixedFluidWarning_description))



from VESIcal import core
from VESIcal import models
from VESIcal import sample_class
from VESIcal import calculate_classes
from VESIcal import batchfile

import numpy as np
import warnings as w
import sys
w.filterwarnings("ignore", message="rubicon.objc.ctypes_patch has only been tested ")

# -------------- MELTS preamble --------------- #
from thermoengine import equilibrate
# instantiate thermoengine equilibrate MELTS instance
melts = equilibrate.MELTSmodel('1.2.0')

# Suppress phases not required in the melts simulation
phases = melts.get_phase_names()
for phase in phases:
    melts.set_phase_inclusion_status({phase: False})
melts.set_phase_inclusion_status({'Fluid': True, 'Liquid': True})
# --------------------------------------------- #

# -------------- BATCH PROCESSING ----------- #
class BatchFile(batchfile.BatchFile):
    """Performs model functions on a batchfile.BatchFile object
    """
    pass

    def get_XH2O_fluid(self, sample, temperature, pressure, H2O, CO2):
        """An internally used function to calculate fluid composition.

        Parameters
        ----------
        sample: dictionary
            Sample composition in wt% oxides

        temperature: float
            Temperature in degrees C.

        pressure: float
            Pressure in bars

        H2O: float
            wt% H2O in the system

        CO2: float
            wt% CO2 in the system

        Returns
        -------
        float
            Mole fraction of H2O in the H2O-CO2 fluid

        """
        pressureMPa = pressure / 10.0

        bulk_comp = {oxide:  sample[oxide] for oxide in core.oxides}
        bulk_comp["H2O"] = H2O
        bulk_comp["CO2"] = CO2
        feasible = melts.set_bulk_composition(bulk_comp)

        output = melts.equilibrate_tp(temperature, pressureMPa, initialize=True)
        (status, temperature, pressureMPa, xmlout) = output[0]
        fluid_comp = melts.get_composition_of_phase(xmlout, phase_name='Fluid', mode='component')
            #NOTE mode='component' returns endmember component keys with values in mol fraction.

        if "Water" in fluid_comp:
            H2O_fl = fluid_comp["Water"]
        else:
            H2O_fl = 0.0

        # if H2O_fl == 0:
        #   raise SaturationError("Composition not fluid saturated.")

        return H2O_fl

    def calculate_dissolved_volatiles(self, temperature, pressure, X_fluid=1, print_status=True, model='MagmaSat', record_errors=False, **kwargs):
        """
        Calculates the amount of H2O and CO2 dissolved in a magma at the given P/T conditions and fluid composition. Fluid composition
        will be matched to within 0.0001 mole fraction.

        Parameters
        ----------
        temperature: float, int, or str
            Temperature, in degrees C. Can be passed as float, in which case the
            passed value is used as the temperature for all samples. Alternatively, temperature information for each individual
            sample may already be present in the BatchFile object. If so, pass the str value corresponding to the column
            title in the BatchFile object.

        presure: float, int, or str
            Pressure, in bars. Can be passed as float or int, in which case the
            passed value is used as the pressure for all samples. Alternatively, pressure information for each individual
            sample may already be present in the BatchFile object. If so, pass the str value corresponding to the column
            title in the BatchFile object.

        X_fluid: float, int, or str
            OPTIONAL: Default value is 1. The mole fraction of H2O in the H2O-CO2 fluid. X_fluid=1 is a pure H2O fluid. X_fluid=0 is a
            pure CO2 fluid. Can be passed as a float or int, in which case the passed value is used as the X_fluid for all samples.
            Alternatively, X_fluid information for each individual sample may already be present in the BatchFile object. If so, pass
            the str value corresponding to the column title in the BatchFile object.

        print_status: bool
            OPTIONAL: The default value is True, in which case the progress of the calculation will be printed to the terminal.
            If set to False, nothing will be printed. MagmaSat calculations tend to be slow, and so a value of True is recommended
            for most use cases.

        model: string
            OPTIONAL: Default is 'MagmaSat'. Any other model name can be passed here.

        record_errors: bool
            OPTIONAL: If True, any errors arising during the calculation will be recorded as a column.

        Returns
        -------
        pandas DataFrame
            Original data passed plus newly calculated values are returned.
        """
        dissolved_data = self.get_data()

        if isinstance(temperature, str):
            file_has_temp = True
            temp_name = temperature
        elif isinstance(temperature, float) or isinstance(temperature, int):
            file_has_temp = False
        else:
            raise core.InputError("temp must be type str or float or int")

        if isinstance(pressure, str):
            file_has_press = True
            press_name = pressure
        elif isinstance(pressure, float) or isinstance(pressure, int):
            file_has_press = False
        else:
            raise core.InputError("pressure must be type str or float or int")

        if isinstance(X_fluid, str):
            file_has_X = True
            X_name = X_fluid
        elif isinstance(X_fluid, float) or isinstance(X_fluid, int):
            file_has_X = False
            if X_fluid != 0 and X_fluid !=1:
                if X_fluid < 0.001 or X_fluid > 0.999:
                    raise core.InputError("X_fluid is calculated to a precision of 0.0001 mole fraction. \
                                     Value for X_fluid must be between 0.0001 and 0.9999.")
        else:
            raise core.InputError("X_fluid must be type str or float or int")

        H2Ovals = []
        CO2vals = []
        warnings = []
        errors = []
        if model in models.get_model_names(model='mixed'):
            for index, row in dissolved_data.iterrows():
                try:
                    if file_has_temp == True:
                        temperature = row[temp_name]

                    if file_has_press == True:
                        pressure = row[press_name]

                    if file_has_X == True:
                        X_fluid = row[X_name]

                    bulk_comp = self.get_sample_composition(index, asSampleClass=True)
                    bulk_comp.set_default_units(self.default_units)
                    calc = calculate_classes.calculate_dissolved_volatiles(sample=bulk_comp, pressure=pressure, temperature=temperature,
                                                                    X_fluid=(X_fluid, 1-X_fluid), model=model,
                                                                    silence_warnings=True, **kwargs)
                    H2Ovals.append(calc.result['H2O_liq'])
                    CO2vals.append(calc.result['CO2_liq'])
                    warnings.append(calc.calib_check)
                    errors.append('')
                except Exception as inst:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    warnings.append('Calculation Failed.')
                    errors.append(sys.exc_info()[0])
            dissolved_data["H2O_liq_VESIcal"] = H2Ovals
            dissolved_data["CO2_liq_VESIcal"] = CO2vals

            if file_has_temp == False:
                dissolved_data["Temperature_C_VESIcal"] = temperature
            if file_has_press == False:
                dissolved_data["Pressure_bars_VESIcal"] = pressure
            if file_has_X == False:
                dissolved_data["X_fluid_input_VESIcal"] = X_fluid
            dissolved_data["Model"] = model
            dissolved_data["Warnings"] = warnings
            if record_errors == True:
                dissolved_data["Errors"] = errors

            return dissolved_data

        elif model == 'MagmaSat' or model.model_type == 'MagmaSat':
            XH2Ovals = []
            XCO2vals = []
            FluidProportionvals = []
            iterno = 0
            for index, row in dissolved_data.iterrows():
                iterno += 1
                if print_status == True:
                    percent = iterno/len(dissolved_data.index)
                    batchfile.status_bar.status_bar(percent, index)

                if file_has_temp == True:
                    temperature = row[temp_name]
                if temperature <= 0:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    XH2Ovals.append(np.nan)
                    XCO2vals.append(np.nan)
                    FluidProportionvals.append(np.nan)
                    warnings.append("Sample skipped. Bad temperature.")
                    errors.append(sys.exc_info()[0])
                    w.warn("Temperature for sample " + str(index) + " is <=0. Skipping sample.", stacklevel=2)

                if file_has_press == True:
                    pressure = row[press_name]
                if temperature >0 and pressure <= 0:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    XH2Ovals.append(np.nan)
                    XCO2vals.append(np.nan)
                    FluidProportionvals.append(np.nan)
                    warnings.append("Sample skipped. Bad pressure.")
                    errors.append(sys.exc_info()[0])
                    w.warn("Pressure for sample " + str(index) + " is <=0. Skipping sample.", stacklevel=2)

                if file_has_X == True:
                    X_fluid = row[X_name]
                if temperature >0 and pressure >0 and X_fluid <0:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    XH2Ovals.append(np.nan)
                    XCO2vals.append(np.nan)
                    FluidProportionvals.append(np.nan)
                    warnings.append("Sample skipped. Bad X_fluid.")
                    errors.append(sys.exc_info()[0])
                    w.warn("X_fluid for sample " + str(index) + " is <0. Skipping sample.", stacklevel=2)

                if temperature >0 and pressure >0 and X_fluid >1:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    XH2Ovals.append(np.nan)
                    XCO2vals.append(np.nan)
                    FluidProportionvals.append(np.nan)
                    warnings.append("Sample skipped. Bad X_fluid.")
                    errors.append(sys.exc_info()[0])
                    w.warn("X_fluid for sample " + str(index) + " is >1. Skipping sample.", stacklevel=2)

                if temperature > 0 and pressure > 0 and X_fluid >=0 and X_fluid <=1:
                    try:
                        bulk_comp = self.get_sample_composition(index, asSampleClass=True)
                        bulk_comp.set_default_units(self.default_units)
                        calc = calculate_classes.calculate_dissolved_volatiles(sample=bulk_comp, pressure=pressure, temperature=temperature,
                                                                        X_fluid=X_fluid, model=model, silence_warnings=True,
                                                                        verbose=True)
                        H2Ovals.append(calc.result['H2O_liq'])
                        CO2vals.append(calc.result['CO2_liq'])
                        XH2Ovals.append(calc.result['XH2O_fl'])
                        XCO2vals.append(calc.result['XCO2_fl'])
                        FluidProportionvals.append(calc.result['FluidProportion_wt'])
                        warnings.append(calc.calib_check)
                        errors.append('')
                    except Exception as inst:
                        H2Ovals.append(np.nan)
                        CO2vals.append(np.nan)
                        XH2Ovals.append(np.nan)
                        XCO2vals.append(np.nan)
                        FluidProportionvals.append(np.nan)
                        warnings.append('Calculation Failed.')
                        errors.append(sys.exc_info()[0])
            dissolved_data["H2O_liq_VESIcal"] = H2Ovals
            dissolved_data["CO2_liq_VESIcal"] = CO2vals

            if file_has_temp == False:
                dissolved_data["Temperature_C_VESIcal"] = temperature
            if file_has_press == False:
                dissolved_data["Pressure_bars_VESIcal"] = pressure
            if file_has_X == False:
                dissolved_data["X_fluid_input_VESIcal"] = X_fluid
            dissolved_data["Model"] = model
            dissolved_data["Warnings"] = warnings
            if record_errors == True:
                dissolved_data["Errors"] = errors

            return dissolved_data
        
        else:
            XH2Ovals = []
            XCO2vals = []
            FluidProportionvals = []
            for index, row in dissolved_data.iterrows():
                if file_has_temp == True:
                    temperature = row[temp_name]
                if file_has_press == True:
                    pressure = row[press_name]
                if file_has_X == True:
                    X_fluid = row[X_name]
                bulk_comp = self.get_sample_composition(index, asSampleClass=True)
                bulk_comp.set_default_units(self.default_units)
                if 'Water' in model:
                    try:
                        calc = calculate_classes.calculate_dissolved_volatiles(sample=bulk_comp, pressure=pressure, temperature=temperature,
                                                             X_fluid=X_fluid, model=model, silence_warnings=True)
                        H2Ovals.append(calc.result)
                        warnings.append(calc.calib_check)
                    except:
                        H2Ovals.append(0)
                        warnings.append('Calculation Failed #001')
                if 'Carbon' in model:
                    try:
                        calc = calculate_classes.calculate_dissolved_volatiles(sample=bulk_comp, pressure=pressure, temperature=temperature,
                                                             X_fluid=X_fluid, model=model, silence_warnings=True)
                        CO2vals.append(calc.result)
                        warnings.append(calc.calib_check)
                    except:
                        CO2vals.append(0)
                        warnings.append('Calculation Failed #002')

            if 'Water' in model:
                dissolved_data["H2O_liq_VESIcal"] = H2Ovals
            if 'Carbon' in model:
                dissolved_data["CO2_liq_VESIcal"] = CO2vals
            if file_has_temp == False:
                dissolved_data["Temperature_C_VESIcal"] = temperature
            if file_has_press == False:
                dissolved_data["Pressure_bars_VESIcal"] = pressure
            if file_has_X == False:
                dissolved_data["X_fluid_input_VESIcal"] = X_fluid
            dissolved_data["Model"] = model
            dissolved_data["Warnings"] = warnings

            return dissolved_data

    def calculate_equilibrium_fluid_comp(self, temperature, pressure=None, print_status=False, model='MagmaSat', **kwargs):
        """
        Returns H2O and CO2 concentrations in wt% or mole fraction in a fluid in equilibrium with the given sample(s) at the given P/T condition.

        Parameters
        ----------
        sample: BatchFile object
            Compositional information on samples in oxides.

        temperature: float, int, or str
            Temperature, in degrees C. Can be passed as float, in which case the
            passed value is used as the temperature for all samples. Alternatively, temperature information for each individual
            sample may already be present in the BatchFile object. If so, pass the str value corresponding to the column
            title in the  BatchFile object.

        presure: float, int, or str
            Pressure, in bars. Can be passed as float or int, in which case the
            passed value is used as the pressure for all samples. Alternatively, pressure information for each individual
            sample may already be present in the BatchFile object. If so, pass the str value corresponding to the column
            title in the BatchFile object.

        model: string
            OPTIONAL: Default is 'MagmaSat'. Any other model name can be passed here.

        Returns
        -------
        pandas DataFrame
            Original data passed plus newly calculated values are returned.
        """
        fluid_data = self.data.copy()

        if isinstance(temperature, str):
            file_has_temp = True
            temp_name = temperature
        elif isinstance(temperature, float) or isinstance(temperature, int):
            file_has_temp = False
        else:
            raise core.InputError("temp must be type str or float or int")

        if isinstance(pressure, str):
            file_has_press = True
            press_name = pressure
        elif isinstance(pressure, float) or isinstance(pressure, int) or type(pressure) == type(None):
            file_has_press = False
        else:
            raise core.InputError("pressure must be type str or float or int")

        H2Ovals = []
        CO2vals = []
        warnings = []
        if model in  models.get_model_names(model='mixed') or model == "MooreWater":
            for index, row in fluid_data.iterrows():
                try:
                    if file_has_temp == True:
                        temperature = row[temp_name]
                    if file_has_press == True:
                        pressure = row[press_name]
                    bulk_comp = sample_class.Sample({oxide:  row[oxide] for oxide in core.oxides}, units='wtpt_oxides')
                    calc = calculate_classes.calculate_equilibrium_fluid_comp(sample=bulk_comp, pressure=pressure, temperature=temperature,
                                                            model=model, silence_warnings=True, **kwargs)

                    H2Ovals.append(calc.result['H2O'])
                    CO2vals.append(calc.result['CO2'])
                    if calc.result['H2O'] == 0 and calc.result['CO2'] == 0:
                        warnings.append(calc.calib_check + "Sample not saturated at these conditions")
                    else:
                        warnings.append(calc.calib_check)
                except:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    warnings.append("Calculation Failed.")
            fluid_data["XH2O_fl_VESIcal"] = H2Ovals
            fluid_data["XCO2_fl_VESIcal"] = CO2vals
            if file_has_temp == False:
                fluid_data["Temperature_C_VESIcal"] = temperature
            if file_has_press == False:
                fluid_data["Pressure_bars_VESIcal"] = pressure
            fluid_data["Model"] = model
            fluid_data["Warnings"] = warnings

            return fluid_data
        elif model == 'MagmaSat' or model.model_type == 'MagmaSat':
            iterno = 0
            for index, row in fluid_data.iterrows():
                iterno += 1
                if print_status == True:
                    percent = iterno/len(fluid_data.index)
                    batchfile.status_bar.status_bar(percent, index)

                if file_has_temp == True:
                    temperature = row[temp_name]
                if temperature <= 0:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    warnings.append("Calculation skipped. Bad temperature.")
                    w.warn("Temperature for sample " + str(index) + " is <=0. Skipping sample.", stacklevel=2)

                if file_has_press == True:
                    pressure = row[press_name]
                if temperature >0 and pressure <= 0:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    warnings.append("Calculation skipped. Bad pressure.")
                    w.warn("Pressure for sample " + str(index) + " is <=0. Skipping sample.", stacklevel=2)

                if temperature > 0 and pressure > 0:
                    try:
                        bulk_comp = sample_class.Sample({oxide:  row[oxide] for oxide in core.oxides}, units='wtpt_oxides')
                        calc = calculate_classes.calculate_equilibrium_fluid_comp(sample=bulk_comp, pressure=pressure, temperature=temperature, model=model, silence_warnings=True)

                        H2Ovals.append(calc.result['H2O'])
                        CO2vals.append(calc.result['CO2'])
                        if calc.result['H2O'] == 0 and calc.result['CO2'] == 0:
                            warnings.append(calc.calib_check + "Sample not saturated at these conditions")
                        else:
                            warnings.append(calc.calib_check)
                    except:
                        H2Ovals.append(np.nan)
                        CO2vals.append(np.nan)
                        warnings.append("Calculation Failed.")

            fluid_data["XH2O_fl_VESIcal"] = H2Ovals
            fluid_data["XCO2_fl_VESIcal"] = CO2vals
            if file_has_temp == False:
                fluid_data["Temperature_C_VESIcal"] = temperature
            if file_has_press == False:
                fluid_data["Pressure_bars_VESIcal"] = pressure
            fluid_data["Model"] = model
            fluid_data["Warnings"] = warnings

            return fluid_data

        else:
            saturated = []
            for index, row in fluid_data.iterrows():
                try:
                    if file_has_temp == True:
                        temperature = row[temp_name]
                    if file_has_press == True:
                        pressure = row[press_name]
                    bulk_comp = sample_class.Sample({oxide:  row[oxide] for oxide in core.oxides}, units='wtpt_oxides')
                    calc = calculate_classes.calculate_equilibrium_fluid_comp(sample=bulk_comp, pressure=pressure, temperature=temperature, model=model, silence_warnings=True)
                    saturated.append(calc.result)
                    warnings.append(calc.calib_check)
                except:
                    saturated.append(np.nan)
                    warnings.append("Calculation Failed.")
            fluid_data["Saturated_VESIcal"] = saturated
            if file_has_temp == False:
                fluid_data["Temperature_C_VESIcal"] = temperature
            if file_has_press == False:
                fluid_data["Pressure_bars_VESIcal"] = pressure
            fluid_data["Model"] = model
            fluid_data["Warnings"] = warnings

            return fluid_data

    def calculate_saturation_pressure(self, temperature, print_status=None, model='MagmaSat', **kwargs):
        """
        Calculates the saturation pressure of multiple sample compositions in the BatchFile.

        Parameters
        ----------
        temperature: float, int, or str
            Temperature at which to calculate saturation pressures, in degrees C. Can be passed as float or int, in which case the
            passed value is used as the temperature for all samples. Alternatively, temperature information for each individual
            sample may already be present in the passed BatchFile object. If so, pass the str value corresponding to the column
            title in the passed BatchFile object.

        print_status: bool
            OPTIONAL: The default value for MagmaSat is True and the default for all other models is False. If set to True, the 
            progress of the calculation will be printed to the terminal. If set to False, nothing will be printed. MagmaSat 
            calculations tend to be slow, and so a value of True is recommended more most use cases.

        model: string
            OPTIONAL: Default is 'MagmaSat'. Any other model name can be passed here.

        Returns
        -------
        pandas DataFrame object
            Values returned are saturation pressure in bars, the mass of fluid present, and the composition of the
            fluid present.
        """
        satp_data = self.data.copy()

        # set default print_status to True for MagmaSat, False for other models if user doesn't pass any option
        if print_status == None:
            if model == 'MagmaSat' or model.model_type == 'MagmaSat':
                print_status = True
            else:
                print_status = False

        if isinstance(temperature, str):
            file_has_temp = True
            temp_name = temperature
        elif isinstance(temperature, float) or isinstance(temperature, int):
            file_has_temp = False
        else:
            raise core.InputError("temperature must be type str or float or int")

        if model != 'MagmaSat':
            satP = []
            warnings = []
            piStar = []
            iterno = 0
            for index, row in satp_data.iterrows():
                iterno += 1
                if print_status == True:
                    percent = iterno/len(satp_data.index)
                    batchfile.status_bar.status_bar(percent, index)
                if file_has_temp == True:
                    temperature = row[temp_name]
                bulk_comp = sample_class.Sample({oxide:  row[oxide] for oxide in core.oxides}, units='wtpt_oxides')
                calc = calculate_classes.calculate_saturation_pressure(sample=bulk_comp, temperature=temperature,
                                                     model=model, silence_warnings=True, **kwargs)
                satP.append(calc.result)
                warnings.append(calc.calib_check)
                if model == 'Shishkina':
                    piStar.append(models.default_models['Shishkina'].models[1].PiStar(bulk_comp))

            satp_data["SaturationP_bars_VESIcal"] = satP
            if file_has_temp == False:
                satp_data["Temperature_C_VESIcal"] = temperature
            satp_data["Model"] = model
            satp_data["Warnings"] = warnings
            if len(piStar)>0:
                satp_data['PiStar_VESIcal'] = piStar

            return satp_data

        elif model == 'MagmaSat' or model.model_type == 'MamgaSat':
            satP = []
            flmass = []
            flH2O = []
            flCO2 = []
            flsystem_wtper = []
            warnings = []
            iterno = 0
            for index, row in satp_data.iterrows():
                iterno += 1
                if print_status == True:
                    percent = iterno/len(satp_data.index)
                    batchfile.status_bar.status_bar(percent, index)

                if file_has_temp == True:
                    temperature = row[temp_name]
                if temperature <= 0:
                    satP.append(np.nan)
                    flmass.append(np.nan)
                    flsystem_wtper.append(np.nan)
                    flH2O.append(np.nan)
                    flCO2.append(np.nan)
                    warnings.append("Calculation skipped. Bad temperature.")
                    w.warn("Temperature for sample " + str(index) + " is <=0. Skipping sample.", stacklevel=2)

                if temperature > 0:
                    try:
                        bulk_comp = sample_class.Sample({oxide:  row[oxide] for oxide in core.oxides}, units='wtpt_oxides')
                        calc = calculate_classes.calculate_saturation_pressure(sample=bulk_comp, temperature=temperature, model=model, verbose=True, silence_warnings=True)
                        satP.append(calc.result["SaturationP_bars"])
                        flmass.append(calc.result["FluidMass_grams"])
                        flsystem_wtper.append(calc.result["FluidProportion_wt"])
                        flH2O.append(calc.result["XH2O_fl"])
                        flCO2.append(calc.result["XCO2_fl"])
                        warnings.append(calc.calib_check)
                    except:
                        satP.append(np.nan)
                        flmass.append(np.nan)
                        flsystem_wtper.append(np.nan)
                        flH2O.append(np.nan)
                        flCO2.append(np.nan)
                        warnings.append("Calculation Failed")

            satp_data["SaturationP_bars_VESIcal"] = satP
            if file_has_temp == False:
                satp_data["Temperature_C_VESIcal"] = temperature
            satp_data["XH2O_fl_VESIcal"] = flH2O
            satp_data["XCO2_fl_VESIcal"] = flCO2
            satp_data["FluidMass_grams_VESIcal"] = flmass
            satp_data["FluidSystem_wt_VESIcal"] = flsystem_wtper
            satp_data["Model"] = model
            satp_data["Warnings"] = warnings

            return satp_data

def BatchFile_from_DataFrame(dataframe, units='wtpt_oxides', label=None):
    """
    Transforms any pandas DataFrame object into a VESIcal BatchFile object.

    Parameters
    ----------
    Same as batchfile.BatchFile_from_DataFrame()

    Returns
    -------
    VESIcal.BatchFile object
    """
    return BatchFile(filename=None, dataframe=dataframe, units=units, label=label)